
from pycamia import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = 'micomputing',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.1.48',
    contact = 'bertiezhou@163.com',
    keywords = ['medical image', 'image registration', 'image similarities'],
    description = "'micomputing' is a package for medical image computing. ",
    requires = ['numpy', 'torch>=1.5.1', 'batorch', 'matplotlib', 'pycamia', 'pyoverload', 'nibabel', 'pydicom', 'SimpleITK'],
    update = '2024-05-22 17:35:23',
    package_data = True
).check()
__version__ = '1.1.48'
from .trans import Transformation, ComposedTransformation, CompoundTransformation, interpolation, interpolation_forward, Identity, Id, Rotation90, Rotation180, Rotation270, Reflect, Reflection, Permutedim, DimPermutation, Rescale, Rescaling, rand_Rescaling, Translate, Translation, rand_Translation, Rigid, Rig, rand_Rigidity, rand_Rig, Affine, Aff, rand_Affinity, rand_Aff, PolyAffine, rand_PolyAffine, logEuclidean, logEu, rand_logEuclidean, rand_logEu, LocallyAffine, LARM, rand_LocallyAffine, rand_LARM, FreeFormDeformation, FFD, rand_FreeFormDeformation, rand_FFD, DenseDisplacementField, DDF, rand_DenseDisplacementField, rand_DDF, VelocityField, VF, rand_VelocityField, rand_VF, Normalize, Cropping #*
from . import plot as plt
from .stdio import IMG, dcm2nii, nii2dcm, affine2orient, affine2spacing, dim_of_orient_axis #*
from .data import Key, DataObject, Slicer, Dataset, LossCollector #*
from .network import U_Net, CNN, FCN, RNN, NeuralODE, Convolution_Block, Convolution, Models #*
from .funcs import reorient, rescale, dilate, blur, bending, distance_map, registration, local_prior, center_of_gravity #*
# from .trans import Transformation, WoldCoordsTransformation, ImageCoordsTransformation, IntensityTransformation, ComposedTransformation, CompoundTransformation, Identity, Id, Rotation90, Rotation180, Rotation270, Reflect, Reflection, Permutedim, DimPermutation, Rescale, Rescaling, Translate, Translation, Rigid, Rig, Affine, Aff, PolyAffine, logEu, LocallyAffine, LARM, FreeFormDeformation, FFD, DenseDisplacementField, DDF, MultiLayerPerception, MLP, Normalize, resample, interpolation, interpolation_forward, Affine2D2Matrix, Quaterns2Matrix, Matrix2Quaterns #*
from .metrics import metric, ITKMetric, ITKLabelMetric, MutualInformation, NormalizedMutualInformation, KLDivergence, CorrelationOfLocalEstimation, NormalizedVectorInformation, Cos2Theta, SumSquaredDifference, MeanSquaredErrors, PeakSignalToNoiseRatio, CrossEntropy, CrossCorrelation, NormalizedCrossCorrelation, StructuralSimilarity, Dice, DiceScore, DiceScoreCoefficient, LabelDice, LabelDiceScore, LabelDiceScoreCoefficient, ITKDiceScore, ITKJaccardCoefficient, ITKVolumeSimilarity, ITKFalsePositive, ITKFalseNegative, ITKHausdorffDistance, ITKMedianSurfaceDistance, ITKAverageSurfaceDistance, ITKDivergenceOfSurfaceDistance, ITKLabelDiceScore, ITKLabelJaccardCoefficient, ITKLabelVolumeSimilarity, ITKLabelFalsePositive, ITKLabelFalseNegative, ITKLabelHausdorffDistance, ITKLabelMedianSurfaceDistance, ITKLabelAverageSurfaceDistance, ITKLabelDivergenceOfSurfaceDistance, LocalNonOrthogonality, RigidProjectionError, joint_hist #*




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "Useful python structures",
    requires = ""
)

__all__ = """
    struct
    odict
""".split()

from typing import Mapping, Iterable

class struct(dict):
    def __setattr__(self, name, value):
        self[name] = value
    def __getattribute__(self, name):
        try: return super().__getattribute__(name)
        except AttributeError: return self[name]
    def __str__(self):
        return "struct {\n" + ',\n'.join(f"    {n} = {v}" for n, v in self.items()) + "\n}"
    
class odict(dict):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1: arg = args[0]
        self = super().__new__(cls, *args, **kwargs)
        if len(args) == 0: self._keys = []
        elif isinstance(arg, Mapping): self._keys = [x[0] for x in arg.items()]
        elif isinstance(arg, Iterable): self._keys = [x[0] for x in arg]
        elif len(kwargs) > 0: self._keys = list(kwargs.keys())
        else: self._keys = []
        return self

    @classmethod
    def __class_getitem__(cls, args: tuple[slice]):
        if not isinstance(args, tuple): args = (args,)
        self = cls()
        for arg in args:
            if not isinstance(arg, slice): raise TypeError("Creating ordered dict by odict[x:y] format accepts only 'slice' objects in subscript. ")
            if arg.step is not None: raise TypeError("No more than one colon is accepted for odict[x:y] format. ")
            self._keys.append(arg.start)
            self[arg.start] = arg.stop
        return self
    
    def __iter__(self):
        for k in self._keys: yield k
    def __setitem__(self, key, value):
        if key not in self._keys:
            self._keys.append(key)
        return super().__setitem__(key, value)
    def update(self, new_dict):
        self._keys.extend([k for k in new_dict.keys() if k not in self._keys])
        return super().update(new_dict)
    def items(self):
        for k in self._keys:
            yield k, self[k]
    def pop(self, key):
        self._keys.remove(key)
        return super().pop(key)
    def index(self, key):
        return self._keys.index(key)
    def key(self, index):
        return self._keys[index]
    def __str__(self):
        return '[' + ', '.join(f"{k!r}: {v!r}" for k, v in self.items()) + ']'




__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of string operations. "
)

__all__ = """
    is_digits
    is_alphas
    is_snakename
    get_digits
    get_alphas
    get_snakename
    get_snakenames
    str_len
    str_slice
    find_all
    enclosed_object
    tokenize
    token_replace
    columns
""".split()

import re, math

def is_digits(str_:str):
    """
    Return whether `str_` characters are all digits. 
    
    Examples::
        >>> print(is_digits("123"), is_digits("12 3"))
        True False
    """
    return all([48 <= ord(x) <= 57 for x in str(str_)])

def is_alphas(str_:str):
    """
    Return whether `str_` characters are all alphabets. 
    
    Examples::
        >>> print(is_alphas("abc"), is_alphas("1bc"))
        True False
    """
    return all([65 <= ord(x) <= 90 or 97 <= ord(x) <= 122 for x in str(str_)])

def is_snakename(str_:str):
    """
    Return whether `str_` is a snake name (i.e. for variable names or file names), whose characters are all alphabets/digits/underlines. 
    
    Examples::
        >>> print(is_snakename("abc2_x"), is_snakename("1bc"), is_snakename("x_ y"))
        True False False
    """
    str_ = str(str_)
    return (65 <= ord(str_[0]) <= 90 or 97 <= ord(str_[0]) <= 122) and all([48 <= ord(x) <= 57 or 65 <= ord(x) <= 90 or ord(x) == 95 or 97 <= ord(x) <= 122 for x in str_])

def get_digits(str_:str):
    """
    Return all digits in `str_`. 
    
    Examples::
        >>> print(repr(get_digits("12 3")), repr(get_digits("12a3")))
        '123' '123'
    """
    return ''.join(x for x in str(str_) if 48 <= ord(x) <= 57)

def get_alphas(str_:str):
    """
    Return all alphabets in `str_`. 
    
    Examples::
        >>> print(repr(get_alphas("12 3")), repr(get_alphas("12a3")))
        '' 'a'
    """
    return ''.join(x for x in str(str_) if 65 <= ord(x) <= 90 or 97 <= ord(x) <= 122)

def get_snakename(str_:str):
    """
    Return all alphabets/digits/underlines characters in `str_` so that it is a snake name (i.e. for variable names or file names). 
    
    Examples::
        >>> print(repr(get_snakename("abc2_x")), repr(get_snakename("1bc")), repr(get_snakename("x_ y:")))
        'abc2_x' 'bc' 'x_y'
    """
    past_prefix = False
    result = ""
    for x in str(str_):
        if (65 <= ord(x) <= 90 or 97 <= ord(x) <= 122):
            result += x
            past_prefix = True
        elif (48 <= ord(x) <= 57 or ord(x) == 95) and past_prefix:
            result += x
        else:
            continue
    return result

def get_snakenames(str_:str):
    """
    Return all alphabets/digits/underlines (snakename) words in `str_`. 
    
    Examples::
        >>> print(repr(get_snakename("abc2_x")), repr(get_snakename("1bc")), repr(get_snakename("x_ y:")))
        ['abc2_x'] ['bc'] ['x', 'y']
    """
    rec = False
    ret = []
    for c in str(str_):
        if not rec and (65 <= ord(c) <= 90 or 97 <= ord(c) <= 122): rec = True; ret.append(c); continue
        if rec and (65 <= ord(c) <= 90 or 97 <= ord(c) <= 122 or 48 <= ord(c) <= 57 or ord(c) == 95): ret[-1] += c; continue
        if rec: rec = False
    return ret
    

def str_len(str_:str, r:int=2):
    """
    Returen the ASCII string length of `str_`. 
    
    Args:
        r: bytes a wide character stands for. 
    
    Examples::
        >>> print(str_len("12"), len("你好"), str_len("你好"))
        2 2 4
    """
    length = 0
    for c in str(str_):
        if ord(u'\u4e00') <= ord(c) <= ord(u'\u9fa5'): length += r
        elif c == '\t': length += 4
        else: length += 1
    return length

def find_all(str_:str, key:str):
    """
    Returen all the starting indices of string `key` in string `str_`. 
    
    Examples::
        >>> find_all("abcaa", 'a')
        [0, 3, 4]
    """
    p, indices = -1, []
    while True:
        p = str(str_).find(key, p + 1)
        if p < 0: break
        indices.append(p)
    return indices

def str_slice(str_:str, indices:list):
    """
    Split the string `str_` by breaks in list `indices`.
    
    Examples::
        >>> str_slice("abcaa", [2,4])
        ["ab", "ca", "a"]
    """
    indices.insert(0, 0); indices.append(len(str(str_)))
    return [str(str_)[indices[i]:indices[i+1]] for i in range(len(indices) - 1)]

def enclosed_object(str_, by="()[]{}$$``''\"\"", start=0):
    """
    Return the first object enclosed with a whole pair of parenthesis in `str_` after index `start`.
    
    Argument by[str]: A string with brackets such as "()[]''", pairs by pairs. 
        Same characters on both sides like quotes are also included. 
    Note: One may also use the old way - a list of 'left char', 'right char' and 'chars for both sides',
        e.g. ["([{", ")]}", "$`'\""], but this is deprecated. 
    
    Examples::
        >>> enclosed_object("function(something inside), something else. ")
        function(something inside)
    """
    str_ = str(str_)
    if isinstance(by, list):
        if len(by) == 3: left, right, both = by
        elif len(by) == 2: left, right = by; both = ""
        elif len(by) == 1: left = ""; right = ""; both = by[0]
        else: raise TypeError(f"Invalid argument `by` of length {len(by)} for function `enclosed_object`, should be 1-3. ")
    elif isinstance(by, str):
        left = ""; right = ""; both = ""
        for i in range(0, len(by), 2):
            if by[i] == by[i+1]: both += by[i]
            else: left += by[i]; right += by[i+1]
        by = [left, right, both]
    else: raise TypeError(f"Invalid argument `by` of type {type(by)} for function `enclosed_object`. ")
    depth = {'all': 0}
    for i in range(start, len(str_)):
        s = str_[i]
        if s in right:
            if depth.get(s, 0) == 0: return str_[start:i]
            assert depth[s] > 0 and depth['all'] > 0
            depth[s] -= 1
            depth['all'] -= 1
            if depth[s] == 0 and depth['all'] == 0: return str_[start:i+1]
        elif s in left:
            r = right[left.index(s)]
            depth.setdefault(r, 0)
            depth[r] += 1
            depth['all'] += 1
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] > 0:
                depth[s] -= 1;
                depth['all'] -= 1
                if depth[s] == 0 and depth['all'] == 0: return str_[start:i+1]
            else: depth[s] += 1; depth['all'] += 1
    raise RuntimeError(f"Cannot find enclosed object from string {repr(str_)}.")

def tokenize(str_:str, sep=[' ', '\n'], by="()[]{}$$``''\"\"", start=0, strip='', keep_empty=True):
    """
    Split the string `str_` by elements in `sep`, but keep enclosed objects not split.
    
    Args:
        sep (list of str's): all strings that seperate the tokens.
        by (str): A string with brackets such as "()[]''", pairs by pairs. 
            Same characters on both sides like quotes are also included. 
        Note: One may also use the old way - a list of 'left char', 'right char' and 'chars for both sides',
            e.g. ["([{", ")]}", "$`'\""], but this is deprecated. 

    Examples::
        >>> tokenize("function(something inside), something else. ")
        ["function(something inside),", "something", "else.", ""]
    """
    str_ = str(str_)
    if strip == True: strip = None
    if strip == False: strip = ''
    if isinstance(sep, str): sep = [sep]
    if isinstance(by, list):
        if len(by) == 3: left, right, both = by
        elif len(by) == 2: left, right = by; both = ""
        elif len(by) == 1: left = ""; right = ""; both = by[0]
        else: raise TypeError(f"Invalid argument `by` of length {len(by)} for function `tokenize`, should be 1-3. ")
    elif isinstance(by, str):
        left = ""; right = ""; both = ""
        for i in range(0, len(by), 2):
            if by[i] == by[i+1]: both += by[i]
            else: left += by[i]; right += by[i+1]
        by = [left, right, both]
    else: raise TypeError(f"Invalid argument `by` of type {type(by)} for function `tokenize`. ")
    depth = {'all': 0}
    class Tokens(list):
        def __new__(cls, *args, **kwargs):
            self = super().__new__(cls, *args, **kwargs)
            self.indices = []
            return self
        def get(self, k, default=None):
            if not default: return self[self.index(k) + 1]
            else:
                try: return self[self.index(k) + 1]
                except ValueError: return default
        def get_all(self, k, default=None):
            if not default: return self[self.index(k) + 1]
            else:
                try: return self[self.index(k) + 1]
                except ValueError: return default
    tokens = Tokens()
    tokens.starting_indices = []
    p = start
    for i in range(start, len(str_)):
        if depth['all'] == 0:
            for x in sep:
                if str_[i:i + len(x)] == x:
                    t = str_[p:i].strip(strip)
                    if keep_empty or t != '': 
                        tokens.append(t)
                        tokens.indices.append((p, i))
                    p = i + len(x)
        elif depth['all'] == 1:
            for x in sep:
                if len(x) != 1 or x not in right: continue
                if left[right.index(x)] in sep and depth[x] == 1 and str_[i] == x:
                    t = str_[p:i].strip(strip)
                    if keep_empty or t != '': 
                        tokens.append(t)
                        tokens.indices.append((p, i))
                    p = i + 1
        s = str_[i]
        both_done = False
        if s in right:
            if depth.get(s, 0) == 0: break
            assert depth[s] > 0 and depth['all'] > 0
            depth[s] -= 1
            depth['all'] -= 1
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] > 0:
                depth[s] -= 1
                depth['all'] -= 1
                both_done = True
        if s in left:
            r = right[left.index(s)]
            depth.setdefault(r, 0)
            depth[r] += 1
            depth['all'] += 1
        elif both_done: pass
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] == 0:
                depth[s] += 1
                depth['all'] += 1
    t = str_[p:].strip(strip)
    if keep_empty or t != '': 
        tokens.append(t)
        tokens.indices.append((p, len(str_)))
    return tokens

def token_replace(str_:str, expr:str, replacement:str, sep=[' ', '\n'], by="()[]{}$$``''\"\"", start=0, strip='', keep_empty=True):
    if not (isinstance(expr, str) or callable(expr)):
        raise TypeError("'token_replace' only accept str or function for the keyword expression. ")
    tokens = tokenize(str_, sep=sep, by=by, start=start, strip=strip, keep_empty=keep_empty)
    output = []
    p = 0
    for i, t in enumerate(tokens):
        if isinstance(expr, str) and t != expr: continue
        if callable(expr) and not expr(t): continue
        output.append(str_[p:tokens.indices[i][0]])
        output.append(replacement)
        p = tokens.indices[i][1]
    output.append(str_[p:])
    return ''.join(output)

def columns(*strs, sep=' | ', line_width = 100):
    strs = list(strs)
    n = len(strs)
    col_width = (line_width - (n-1) * len(sep)) // n
    strs_lines = [str(s).split('\n') for s in strs]
    output = ""
    for L in range(max([len(str_lines) for str_lines in strs_lines])):
        lines = []
        for str_lines in strs_lines:
            str_line = str_lines[L] if L < len(str_lines) else ''
            column = []
            if len(str_line) > col_width:
                for p in range(math.ceil(len(str_line) / col_width)):
                    column.append(str_line[p*col_width:(p+1)*col_width])
            else: column.append(str_line)
            lines.append(column)
        for i in range(max(len(c) for c in lines)):
            output += sep.join((c[i] + ' ' * (col_width - len(c[i]))) if i < len(c) else ' ' * col_width for c in lines) + '\n'
    return output





from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2022-02",
    fileinfo = "File to log running infos. "
)

__all__ = """
    logging
    start_logging
""".split()

import re, os, sys, math
import builtins
import traceback as tb
from functools import wraps
from datetime import datetime
from .exception import avouch
from .environment import get_environ_vars, _get_frames, stack_error
from .system import Path

logger = None
builtin_print = print
    
def start_logging(file_path=None):
    logger = logging(file_path, _vars = get_environ_vars())
    logger.__enter__()
    builtins.print = logger.print
    client_frame = _get_frames(2)
    func_name = 'start_logging'
    with open(client_frame.f_code.co_filename) as fp:
        for _ in range(client_frame.f_lineno-1): fp.readline()
        l = fp.readline()
        if func_name not in l: raise stack_error

        prefix = re.search(r'^([ \t]*)[^ \t]', l).group(1)
        log_var = l[len(prefix):l.index(func_name)].strip(' =')
        if '\t' in prefix: indent = '\t'
        elif len(prefix) % 4 == 0: indent = '    '
        else: indent = '  '
        code = ""
        while True:
            l = fp.readline()
            if l == '': break
            avouch(l.startswith(prefix))
            l = indent + l[len(prefix):]
            code += l

        code = "try:\n" + code + """
except Exception as e:
    exc_type, exc_value, traceback = sys.exc_info()
    logger.print("Traceback (most recent call last):")
    for item in tb.StackSummary.from_list(tb.extract_tb(traceback)).format():
        logger.print(item, end="")
    logger.print(exc_type.__name__ + ':', exc_value)
    logger.fp.close()
finally:
    os._exit(0)
        """
        locals().update(client_frame.f_locals)
        globals().update(client_frame.f_globals)
        if log_var: exec(f"{log_var} = logger")
        exec(code)
    
def reformat(string, line_width = 72):
    if line_width is None: return string
    return '\n'.join([
        '\n'.join([
            l[i * line_width:(i+1) * line_width]
            for i in range(math.ceil(len(l) / line_width))
        ]) for l in string.split('\n')
    ])
    
class logging:
    def __init__(self, log_path=None, log_dir=None, prefix="{time}", _vars=None, exc_remove=True, line_width=100):
        if isinstance(prefix, str): self.prefix = lambda: prefix.format(time=datetime.now())
        else: self.prefix = prefix
        self.file_path = log_path
        self.file_dir = log_dir
        self.vars = _vars
        self.fp = None
        self.is_fplinestart = True
        self.exc_remove = exc_remove
        self.line_width = line_width
        self.start_time = datetime.now()

    def __enter__(self):
        if self.file_path is not None and not self.file_path: return
        if self.file_dir is not None and not self.file_dir: return
        if self.vars is None: self.vars = get_environ_vars()
        vars = self.vars
        if self.file_path is None:
            self.file_path = Path(vars['__file__']).with_ext('log')
        self.file_path = Path(self.file_path)
        if not self.file_path | 'log':
            self.file_path = self.file_path // 'log'
        time_str = datetime.now().strftime("%Y%m%d%H%M%S")
        self.file_path ^= time_str
        if self.file_dir is not None:
            self.file_path = Path(self.file_dir).mkdir() / self.file_path.filename
        open(self.file_path, 'w').close()
        self.fp = open(self.file_path, 'a')
        builtins.print = self.print
        fill1 = fill2 = ''
        if '__info__' in vars:
            fill1 = f" in {vars['__info__'].project}{'.' + vars['__info__'].package if vars['__info__'].package else ''} ({vars['__info__'].author})"
            fill2 = '\n| ' + vars['__info__'].fileinfo
        header = '=' * self.line_width + f"""
Logging file for python script{fill1}: {Path(vars['__file__']).filename}{fill2}
Started at {datetime.now()} with command:
> python {Path(sys.argv[0]).filename} {' '.join(sys.argv[1:])}
"""
        # if '__info__' in vars:
        header += '=' * self.line_width + "\n\n"
        header = reformat(header, line_width=self.line_width)
        builtin_print(header, end='')
        self.fp.write(header)
        self.fp.flush()
        return self
        
    def __exit__(self, exc_type, exc_value, traceback):
        if self.fp is None: return
        if exc_type is not None:
            exc_string = "Traceback (most recent call last): \n"
            for item in tb.StackSummary.from_list(tb.extract_tb(traceback)).format(): exc_string += str(item)
            exc_string += exc_type.__name__ + ': ' + str(exc_value) + '\n'
            self.fp_print(exc_string)
            self.fp.close()
            if self.exc_remove:
                log_file = Path(self.file_path)
                if "KeyboardInterrupt: " in exc_string: log_file.rename(log_file^"interrupted")
                else: log_file.remove()
                er_file = log_file.with_name("error_msgs")
                if er_file.exists(): mode = 'a'
                else: mode = 'w'
                with er_file.open(mode) as fp: fp.write(('\n' if mode == 'a' else '') + f"[{self.start_time} - {datetime.now()}]: {exc_string}")
        self.fp = None
        builtins.print = builtin_print
        
    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.__enter__()
            try: func(*args, **kwargs)
            except Exception as e:
                self.__exit__(e.__class__, e.__str__, e.__traceback__)
            self.__exit__(None, None, None)
        return wrapper
        
    def print(self, *values, sep=' ', end='\n', **kwargs):
        builtin_print(self.fp_print(*values, sep=sep, end=end, **kwargs), end='')

    def fp_print(self, *values, sep=' ', end='\n', **kwargs):
        line_prefix = self.prefix()
        if "{time}" in line_prefix: line_prefix = line_prefix.replace("{time}", str(datetime.now()))
        blank_prefix = ' ' * len(line_prefix)
        if line_prefix:
            line_prefix += ' | '
            blank_prefix += ' | '
        output_str = sep.join([str(v) for v in values if str(v)]) + \
                     sep.join([f"{k} = {str(v)}," for k, v in kwargs.items() if str(v)]) + end
        output_str = reformat(output_str, line_width=self.line_width)
        first_line, *other_lines = output_str.split('\n')
        if len(other_lines) == 0:
            self.fp.write((line_prefix if self.is_fplinestart else '') + first_line)
            self.is_fplinestart = False
        else:
            last_line = other_lines.pop(-1)
            self.fp.write((line_prefix if self.is_fplinestart else '') + first_line + '\n')
            self.fp.write(''.join([blank_prefix + l + '\n' for l in other_lines]))
            if last_line == '': self.is_fplinestart = True
            else: self.fp.write(blank_prefix + last_line); self.is_fplinestart = False
        self.fp.flush()
        return output_str

    def std_print(self, *values, sep=' ', end='\n'):
        builtin_print(*values, sep=sep, end=end)




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to handle the outputs.",
    help = "Use `sprint = SPrint()` to start printing to a string, etc. "
)

__all__ = """
    no_out
    no_print
    SPrint

    StrIO
    input_with_timeout
""".split()

import os, sys

class StrIO:
    def __init__(self, file_name = os.path.abspath('.null')):
        self._str_ = None
        self._file_ = open(file_name, 'w+')
        self.file_name = file_name
        self.fileno = self._file_.fileno
    def write(self, s): self._file_.write(s)
    def __str__(self):
        if self._str_ is not None: return self._str_
        self._file_.seek(0)
        self._str_ = self._file_.read()
        self.close()
        return self._str_
    def split(self, c=None): return str(self).split(c)
    def string(self): return str(self)
    def close(self):
        self._file_.close()
        if self.file_name == os.path.abspath('.null'):
            os.remove(self.file_name)
            self._file_ = None

class NPrint:
    """
    Suppress the outputs, designed for instance `no_print`.
    """
    def __init__(self, no_error=False):
        self.no_error = no_error
    def __enter__(self):
        self.io = StrIO()
        self.num_out = sys.stdout.fileno()
        if self.no_error: self.num_err = sys.stderr.fileno()
        self.old_out = os.fdopen(os.dup(sys.stdout.fileno()), 'wb')
        if self.no_error: self.old_err = os.fdopen(os.dup(sys.stderr.fileno()), 'wb')
        os.dup2(self.io.fileno(), self.num_out)
        if self.no_error: os.dup2(self.io.fileno(), self.num_err)
        return self.io
    def __exit__(self, exc_type, exc_value, traceback):
        os.dup2(self.old_out.fileno(), self.num_out)
        if self.no_error: os.dup2(self.old_err.fileno(), self.num_err)
        self.io.string()

no_out = NPrint()
no_print = NPrint(no_error=True)
"""
Suppress the outputs.

Examples::
    >>> print("something in the front")
    >>> with no_print as io:
    >>>     print("something in the middle")
    >>> print("something in behind")
    >>> print(io.text)
    something in the front
    something in behind
    something in the middle
"""

class SPrint:
    """
    Print to a string.
    
    Args:
        init_text (str): initial text, as the head of the resulting string.
        sep (str): the seperator between different elements, just like the built-in function `print`.
        end (str): the ending string of each output, just like the built-in function `print`.

    Examples::
        >>> output = SPrint("!>> ")
        >>> output("Use it", "like", 'the function', "'print'.", sep=' ')
        !>> Use it like the function 'print'.
        >>> output("A return is added automatically each time", end=".")
        !>> Use it like the function 'print'.
        A return is added automatically each time.
        >>> output.text
        !>> Use it like the function 'print'.
        A return is added automatically each time.
    """

    def __init__(self, init_text='', sep=' ', end='\n', print_onscreen=False):
        self._text = init_text
        self.def_sep = sep
        self.def_end = end
        self.def_print = print_onscreen

    def __call__(self, *parts, sep=None, end=None, print_onscreen=None):
        if sep is None: sep = self.def_sep
        if end is None: end = self.def_end
        if print_onscreen is None: print_onscreen = self.def_print
        line = sep.join([str(x) for x in parts if str(x)])
        if print_onscreen: print(line, end=end)
        self._text += line + end
        return self._text

    def __str__(self): return self.text
    
    @property
    def text(self):
        t = self._text
        self._text = ""
        return t

    def clear(self): self._text = ""
    
    def save(self, file_path):
        with open(file_path, 'w') as fp:
            fp.write(self._text)
            
    def append_to(self, file_path):
        with open(file_path, 'a') as fp:
            fp.write(self._text)

SP = ' '
CR = '\r'
LF = '\n'
CRLF = CR + LF

class TimeoutOccurred(Exception): ...

def echo(string):
    sys.stdout.write(string)
    sys.stdout.flush()

def posix_inputimeout(prompt='', timeout=60):
    echo(prompt)
    sel = selectors.DefaultSelector()
    sel.register(sys.stdin, selectors.EVENT_READ)
    events = sel.select(timeout)

    if events:
        key, _ = events[0]
        return key.fileobj.readline().rstrip(LF)
    else:
        echo(LF)
        termios.tcflush(sys.stdin, termios.TCIFLUSH)
        raise TimeoutOccurred

def win_inputimeout(prompt='', timeout=60):
    echo(prompt)
    begin = time.monotonic()
    end = begin + timeout
    line = ''

    while time.monotonic() < end:
        if msvcrt.kbhit():
            c = msvcrt.getwche()
            if c in (CR, LF):
                echo(CRLF)
                return line
            if c == '\003':
                raise KeyboardInterrupt
            if c == '\b':
                line = line[:-1]
                cover = SP * len(prompt + line + SP)
                echo(''.join([CR, cover, CR, prompt, line]))
            else:
                line += c
        time.sleep(0.05)

    echo(CRLF)
    raise TimeoutOccurred

try: import msvcrt
except ImportError:
    import selectors
    import termios
    input_with_timeout = posix_inputimeout
else:
    import time
    input_with_timeout = win_inputimeout




try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "Types for typehint and overload. ",
    requires = ""
)

__all__ = """
    as_type
    get_type_name
    ArrayType

    class_ class_type
    dtype
    union
    intersection intersect insct
    avoid
    tag
    note

    class_satisfies type_satisfies
    instance_satisfies
    
    object
    bool
    short
    int
    long
    half
    float
    double
    rational
    real
    complex
    number
    scalar
    property
    slice
    null
    
    callable
    functional
    lambda_ lambda_func
    method
    function
    builtin_function_or_method
    method_descriptor
    method_wrapper
    generator_function
    classmethod
    staticmethod
    
    str
    bytearray
    bytes
    memoryview
    map
    filter
    
    array
    iterable
    sequence
    list
    tuple
    dict
    set
    reversed
    frozenset

    range
    generator
    
    zip
    enumerate
""".split()

import re, sys, random, copy, builtins
with __info__:
    from pycamia import alias, avouch, crashed

### >>> basic type functions <<< ###

def _touch(f, d=None):
    try: return f()
    except Exception: return d

def as_type(t: (type, tuple, str, None)):
    if isinstance(t, builtins.type): return t
    elif isinstance(t, builtins.tuple): return union(*t)
    elif isinstance(t, builtins.str): return tag(t)
    elif t is None: return builtins.object
    else: raise TypeError(f"Invalid 'type' {repr(t)}. ")

def get_type_name(x: (type, tuple, str, None)):
    if isinstance(x, builtins.type): return x.__name__
    elif isinstance(x, builtins.str): return repr(x)
    elif isinstance(x, builtins.tuple): return repr(builtins.tuple(get_type_name(i) for i in x))
    elif isinstance(x, builtins.slice): return f"{get_type_name(x.start)}:{get_type_name(x.stop)}"
    elif x is ...: return '...'
    elif x is None: return '-'
    else: raise TypeError(f"Invalid 'type' {repr(x)}. ")

### >>> basic type classes <<< ###

@alias("class_", "class_type")
class type(builtins.type):

    @builtins.staticmethod
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and not kwargs: return args[0].__class__
        elif len(args) == 0 and 'name' not in kwargs or len(args) > 3:
            raise TypeError("type() takes 1 or 3 arguments")
        t_name = args[0] if len(args) >= 1 else kwargs['name']
        bases = args[1] if len(args) >= 2 else kwargs.pop('bases', (builtins.object,))
        dict = args[2] if len(args) >= 3 else (kwargs['dict'] if 'dict' in kwargs else kwargs)
        parent = None
        if not isinstance(bases, builtins.tuple):
            parent = bases
            bases = bases.__mro__
        elif len(bases) == 0 or bases[-1] != builtins.object: bases = bases + (builtins.object,)
        if bases[0].__class__ != builtins.type: bases = (builtins.object,)
        # dict may have 'is_iterable' for iterability test.
        self = super().__new__(cls, t_name, bases, dict)
        if 'is_iterable' not in dict: self.is_iterable = False
        self.properties = dict
        self.parent = dict.get('parent', parent)
        self.locked = False
        return self
    
    def __init__(self, *args, **kwargs): ...

    def __getitem__(self, index: '(list, tuple[(type, slice, int, object)])') -> 'pyoverlaod.ArrayType':
        if self.locked: raise self.lock_error
        if '__getitem__' in self.properties: return self.properties['__getitem__'](self, index)
        if isinstance(index, builtins.list):
            index_names = ', '.join(get_type_name(t) for t in index)
            array_name = f"{get_type_name(self)}[{len(index)}]:[{index_names}]"
            avouch(self.is_iterable, f"Invalid subscripting {self.__name__} with list subscript for non-iterable object.")
            avouch(all(isinstance(x, builtins.type) for x in index if x != ...), f"types list for {array_name} contains non-type element. ")
            return ArrayType(array_name, self, element_types=index)
        if not isinstance(index, builtins.tuple): index = (index,)
        base_type = []
        base_size = None
        type_names = []
        for i in index:
            if isinstance(i, (builtins.type, builtins.tuple, builtins.str)):
                base_type.append(as_type(i))
                type_names.append(get_type_name(i))
            elif i is None:
                base_type.append(builtins.object)
                type_names.append('-')
            elif isinstance(i, builtins.slice):
                avouch(i.step is None, f"Invalid subscripting {self.__name__}[..., {get_type_name(i.start)}:{get_type_name(i.step)}:{get_type_name(i.stop)}, ...] with two colons in ArrayType subscript.")
                base_type.append(as_type(i.start))
                base_type.append(as_type(i.stop))
                type_names.append(get_type_name(i))
            else: break
        base_type = builtins.tuple(base_type)
        if len(base_type) == 0: base_type = None
        base_size = index[len(type_names):] if len(index) > len(type_names) else None
        size_name = '' if base_size is None else (f"[{base_size[0]}]" if len(base_size) == 1 else ('[-]' if len(base_size) == 0 else f"[{repr(base_size).strip('()')}]"))
        size_name = size_name.replace("Ellipsis", '...')
        if not self.is_iterable:
            avouch(len(type_names) == 0, 
                   f"cannot define the type of elements in '{self.__name__}', one can only use {self.__name__}[{{a size}}] to create iterable type with elements of type {self.__name__}. " if hasattr(getattr(self, 'parent', None), '__len__') else
                   f"type '{self.__name__}' is not iterable, one can only use {self.__name__}[{{a size}}] to create iterable type. ")
            return ArrayType(f"{self.__name__}{size_name}", iterable, base_type=self, base_size=base_size)
        if isinstance(self, ArrayType):
            origin_type = self.origin_type
            if base_type is not None: avouch(self.base_type is None, f"{self.__name__}[{repr(index)}] has conflict in element type: multiple types defined by two level subscripts.")
            elif self.base_type is not None:
                base_type = self.base_type
                type_names = [get_type_name(t) for t in base_type]
            if self.base_size is not None:
                if not self.can_add_size: raise TypeError(f"Compound ArrayType {self.__name__} can not accept more size information in subscript, please ensure that only size of 1 dimension in each pair of brackets")
                if base_size is not None:
                    if len(self.base_size) > 0 and len(base_size) > 1: raise TypeError(f"Please add size of one dimension a time, adding {len(base_size)} at a time is invalid: {self.__name__}[{repr(base_size).strip('()')}]")
                    if len(base_size) == 1 and base_size[0] == ...: base_size = (-1,)
                    if len(self.base_size) == 1 and self.base_size[0] == ...: base_size = (-1,) + base_size
                    else: base_size = self.base_size + base_size
                else: base_size = self.base_size
                size_name = f"[{base_size[0]}]" if len(base_size) == 1 else ('[-]' if len(base_size) == 0 else f"[{repr(base_size).strip('()')}]")
                size_name = size_name.replace("Ellipsis", '...')
        else: origin_type = self
        type_names = ': '.join(type_names)
        if not type_names: type_names = ''
        else: type_names = f'<{type_names}>'
        array_name = f"{get_type_name(origin_type)}{type_names}{size_name}"
        return ArrayType(array_name, origin_type, base_type=base_type, base_size=base_size, 
                         can_add_size=self.can_add_size if isinstance(self, ArrayType) else (base_size is None or len(base_size) <= 1))

    def __lshift__(self, ele_type): return self[ele_type].with_lock(True)
    def __rshift__(self, size):
        if self.locked:
            self.locked = False
            # return self[*size]
            return self[size[0] if hatattr(size, '__len__') and len(size) == 1 else size]
        else: raise TypeError("ending tag '>>' before starting '<<'. ")

    def __invert__(self): return avoid(self)
    @alias('__rand__', '__iand__', '__mul__', '__rmul__', '__imul__')
    def __and__(self, other): return intersection(self, other)
    @alias('__ror__', '__ior__', '__add__', '__radd__', '__iadd__')
    def __or__(self, other): return union(self, other)
    @alias('__itruediv__', '__sub__', '__isub__')
    def __truediv__(self, other): return self & ~other
    @alias('__rsub__')
    def __rtruediv__(self, other): return other & ~self

    def __instancecheck__(self, x):
        if self.locked: raise self.lock_error
        if '__subclasshook__' in self.properties: return self.properties['__subclasshook__'](self, x.__class__)
        if '__instancecheck__' in self.properties: return self.properties['__instancecheck__'](self, x)
        if '__subclasscheck__' in self.properties: return self.properties['__subclasscheck__'](self, x.__class__)
        dtype_correct = False
        if not isinstance(x, builtins.type) and hasattr(x, 'shape') and hasattr(x, 'dtype') and len(x.shape) == 0:
            dtype_correct = issubclass(dtype(x.dtype), dtype(getattr(self, '__dtype__', self)))
        if self.parent is not None: return isinstance(x, self.parent) or x.__class__ == self or dtype_correct
        else: return dtype_correct
        return self in x.__class__.__mro__

    def __subclasscheck__(self, t):
        if self.locked: raise self.lock_error
        if '__subclasshook__' in self.properties: return self.properties['__subclasshook__'](self, t)
        if '__subclasscheck__' in self.properties: return self.properties['__subclasscheck__'](self, t)
        avouch(isinstance(t, builtins.type) or isinstance(t, dtype), f"Non type input for subclass check for {self.__name__}: {t}.")
        dtype_correct = False
        if isinstance(t, dtype): dtype_correct = issubclass(t, dtype(getattr(self, '__dtype__', self)))
        if self.parent is not None: return issubclass(t, self.parent) or t == self or dtype_correct
        else: return dtype_correct
        print(f"Waring: performing brutal check of subclass 'issubclass({t.__name__}, {self.__name__})' as type {self.__name__} does not have implementation '__subclasshook__' or '__subclasscheck__'. ")
        return self in t.__mro__

    def __call__(self, *args, **kwargs):
        if self.locked: raise self.lock_error
        if '__call__' in self.properties: return self.properties['__call__'](self, *args, **kwargs)
        if self.parent is not None: return self.parent(*args, **kwargs)
        if len(self.__mro__) > 2: self.__mro__[1](*args, **kwargs)
        raise NotImplementedError(f"Trying to call type '{self.__name__}', no implementation is found.")
    
    def __eq__(self, other):
        if not isinstance(other, builtins.type): return False
        if len(self.__mro__) != len(other.__mro__): return False
        return all(x.__qualname__ == y.__qualname__ for x, y in builtins.zip(self.__mro__[::-1], other.__mro__[::-1]))
    
    def __hash__(self):
        return super().__hash__() if not hasattr(self, 'module') else hash(f"{self.module + '.' if self.module is not None else ''}{self.name}{self.bits}")
    
    def __setattr__(self, name, value):
        builtins.type.__setattr__(self, name, value)
        if name in ('module', 'name', 'bits', 'bytes'):
            self.__name__ = f"dtype[{self.module + '.' if self.module is not None else ''}{self.name}{self.bits}]"
            self.__qualname__ = '.'.join(self.__qualname__.split('.')[:-1] + [self.__name__])

def create_spec_dtype(self, arg):

    if self.name: raise TypeError(f"Specified '{self.__name__}' object is not callable")

    if isinstance(arg, builtins.str):
        module, *_, _dtype = arg.split('.') if '.' in arg else [None, arg]
    elif isinstance(arg, builtins.type):
        if any(getattr(t, '__module__', '').startswith('pyoverload') for t in arg.__mro__):
            module = None
            _dtype = getattr(arg, '__dtype__', arg)
        elif any(getattr(t, '__module__', '').startswith('builtins') for t in arg.__mro__ if t != builtins.object):
            module = None
            _dtype = arg.__name__.split('.')[-1]
        else:
            module = getattr(arg, '__module__', None)
            _dtype = arg.__name__.split('.')[-1]
    elif hasattr(arg, 'dtype') and isinstance(arg.dtype, dtype):
        module = getattr(arg.__class__, '__module__', None)
        _dtype = getattr(arg, 'dtype')
    elif isinstance(arg, dtype):
        module = getattr(arg.__class__, '__module__', None)
        _dtype = arg
    else: raise TypeError(f"Unrecognized dtype '{arg}' (of type {builtins.type(arg)})")
    
    if isinstance(_dtype, type): _dtype = _dtype.__name__
    elif hasattr(_dtype, 'name'): _dtype = _dtype.name
    elif isinstance(_dtype, str): ...
    else: _dtype = str(_dtype)
    if _dtype.endswith('Tensor'): _dtype = _dtype[:-len('Tensor')].lower()
    if not _dtype: return dtype
    
    type_infos = re.findall(r"([<a-zA-Z]+)([x0-9]*)", _dtype.split('.')[-1])
    avouch(len(type_infos) == 1, TypeError(f"Unrecognized dtype '{_dtype}'"))
    name, bits = type_infos[0]
    bits = _touch(lambda: int(bits), bits)
    
    special_dict = builtins.dict(short=('int', 16), half=('float', 16), long=('int', 64), double=('float', 64))
    if name in special_dict:
        avouch(not bits, TypeError(f"Unrecognized dtype '{arg}': {name}{bits}"))
        name, bits = special_dict[name]

    return type.__new__(dtype, f"dtype[{module + '.' if module is not None else ''}{name}{bits}]", 
        module = module.split('.')[0] if module is not None else module, name = name, bits = bits, bytes=bits // 8 if isinstance(bits, int) else (bits + '/8' if isinstance(bits, str) else (bits, 8)), 
    )

def __dtype_subclasscheck__(self, t):
    if dtype not in t.__class__.__mro__: t = dtype(t)
    if self.module is not None and t.module != self.module: return False
    if self.bits is not None and t.bits != self.bits: return False
    return self.name in t.name

dtype = type("dtype", type,
    module = None, name = '', bits = 0, bytes = 0, 
    __call__ = create_spec_dtype,
    __instancecheck__ = lambda self, x: isinstance(x, scalar) and isinstance(x, self[...]),
    __subclasscheck__ = __dtype_subclasscheck__
)
dtype.properties['__instancecheck__'] = lambda self, x: x.__class__ == self or any('dtype' in mt.__name__ for mt in x.__class__.__mro__)
dtype.properties['__subclasscheck__'] = lambda self, t: isinstance(t, builtins.type) and any('dtype' in mt.__name__ for mt in t.__mro__)

class ArrayType(type):
    
    @builtins.staticmethod
    def __new__(cls, array_name, origin_type, **kwargs: 'dict(element_types=tuple[type], base_type=(null, tuple[type]), base_size=(null, tuple[(int, object)]), can_add_size=bool)'):
        avouch(
            'element_types' in kwargs or 'base_type' in kwargs and 'base_size' in kwargs,
            f"Initialization of 'ArrayType' takes either keyword argument 'element_types' or keyword arguments 'base_type' and 'base_size'."
        )
        if 'element_types' in kwargs:
            return super().__new__(cls, array_name, origin_type.__mro__, is_iterable=True, origin_type=origin_type, element_types=kwargs.get('element_types'), can_add_size = False)
        
        base_type = kwargs.get('base_type')
        base_size = kwargs.get('base_size')
        can_add_size = kwargs.get('can_add_size')
        if base_type is not None and not isinstance(base_type, builtins.tuple): base_type = (base_type,)
        avouch(base_type is None or len(base_type) <= 1 or
            len(base_type) == 2 and issubclass(origin_type, builtins.dict) or
            issubclass(origin_type, builtins.zip),
            f"Invalid ArrayType: {array_name}; at most 2 types for dict key-value pair or multiple types for 'zip' in subscript. One should use 'union' or tuple of types to include multiple probabilities, or use a list inside subscript to list the types for each item: e.g. list[[int, float, ...]]. \n" + 
            "P.S. Avoid using a single tuple of types such as 'list[(int, float)]', use 'union' instead or add another comma: 'list[(int, float),]'. "
        )
        if can_add_size is None: can_add_size = base_size is None or len(base_size) <= 1
        return super().__new__(cls, array_name, origin_type.__mro__, is_iterable=True, origin_type=origin_type, base_type=base_type, base_size=base_size, can_add_size=can_add_size)
    
    def __init__(self, array_name, origin_type, **kwargs): ...

    def __instancecheck__(self, x):
        if self.locked: raise self.lock_error
        if not isinstance(x, self.origin_type): return False
        
        if hasattr(self, 'element_types'):
            y = builtins.list(x)
            if ... in self.element_types:
                if len(self.element_types) <= len(y) + 1:
                    i_ellipsis = [i for i, t in enumerate(self.element_types) if t == ...]
                    if len(i_ellipsis) > 1: return False
                    return (
                        all(isinstance(u, t) for u, t in builtins.zip(y[:i_ellipsis[0]], self.element_types[:i_ellipsis[0]])) and 
                        all(isinstance(u, t) for u, t in builtins.zip(y[len(y) - len(self.element_types) + i_ellipsis[0] + 1:], self.element_types[i_ellipsis[0]+1:]))
                    )
                return False
            if len(y) != len(self.element_types):
                i_ellipsis = [i for i, t in enumerate(self.element_types) if getattr(t, 'base_size', None) == (...,)]
                if len(i_ellipsis) != 1: return False
                mid_t = self.element_types[i_ellipsis[0]]
                if mid_t.origin_type == iterable:
                    mid_t = builtins.tuple(mid_t.base_type)
                else: mid_t = mid_t.origin_type
                return (
                    all(isinstance(u, t) for u, t in builtins.zip(y[:i_ellipsis[0]], self.element_types[:i_ellipsis[0]])) and 
                    all(isinstance(u, t) for u, t in builtins.zip(y[len(y) - len(self.element_types) + i_ellipsis[0] + 1:], self.element_types[i_ellipsis[0]+1:])) and 
                    all(isinstance(u, mid_t) for u in y[i_ellipsis[0]:len(y) - len(self.element_types) + i_ellipsis[0] + 1])
                )
            return all(isinstance(u, t) for u, t in builtins.zip(y, self.element_types))
        n_max_ele = 50
        if issubclass(self.origin_type, builtins.dict):
            if self.base_size is not None:
                if len(self.base_size) != 1: raise TypeError(f"in {self}: dict size should be a single integer of total number of elements")
                if self.base_size[0] >= 0 and len(x) != self.base_size[0]: return False
            if self.base_type is None: self.base_type = (builtins.object, builtins.object)
            elif len(self.base_type) == 1: self.base_type = (builtins.object, self.base_type[0])
            if len(x) > n_max_ele: n_keys = n_max_ele
            else: n_keys = len(x)
            return all(isinstance(k, self.base_type[0]) and isinstance(x[k], self.base_type[1]) for k, _ in builtins.zip(x, builtins.range(n_max_ele)))
        generator_types = {
            builtins.zip: lambda x, t: all(isinstance(u, v) for u, v in builtins.zip(x, t)),
            builtins.enumerate: lambda x, t: isinstance(x[1], t[0]),
            generator: lambda x, t: isinstance(x, t) or isinstance(x, tuple) and isinstance(t, tuple) and all(isinstance(u, v) for u, v in builtins.zip(x, t))
        }
        for generator_type, check_generated in generator_types.items():
            if issubclass(self.origin_type, generator_type):
                if self.base_size is not None:
                    if len(self.base_size) != 1: raise TypeError(f"in {self}: {generator_type.__name__} size should be a single integer of total number of elements")
                    if self.base_type is None:
                        try: y = copy.deepcopy(x)
                        except TypeError: raise TypeError(f"Cannot copy the generator, pleas cast it to a list before checking its size. ")
                        if len(builtins.list(y)) != self.base_size[0]: return False
                        return True
                elif self.base_type is None: return True
                try: y = copy.deepcopy(x)
                except TypeError: raise TypeError(f"Cannot copy the generator, pleas cast it to a list before checking its value type. ")
                l = 0
                while True if self.base_size is not None and self.base_size[0] >= 0 else l < n_max_ele:
                    try: sample = next(y)
                    except StopIteration: break
                    if not check_generated(sample, self.base_type): return False
                    l += 1
                if self.base_size is not None and self.base_size[0] >= 0 and self.base_size[0] != l: return False
                return True
        if hasattr(x, 'shape'):
            if self.base_size is not None:
                tests = []
                if ... in self.base_size:
                    i_ellipsis = [i for i, x in enumerate(self.base_size) if x == ...]
                    if len(i_ellipsis) > 1: raise TypeError(f"Only one ellipsis (...) is allowed in size: {repr(self.base_size).replace('Ellipsis', '...')}")
                    tests.append((x.shape[:i_ellipsis[0]], self.base_size[:i_ellipsis[0]]))
                    tests.append((x.shape[len(x.shape)-len(self.base_size)+i_ellipsis[0]+1:], self.base_size[i_ellipsis[0]+1:]))
                else: tests.append((x.shape, self.base_size))
                for a, b in tests:
                    if builtins.type(a)(b) == a: ...
                    elif _touch(lambda: b == builtins.tuple(a)): ...
                    elif len(a) == len(b) and _touch(lambda: all(x == y or y < 0 for x, y in builtins.zip(a, b))): ...
                    else: return False
            if self.base_type is None: return True
            _dtype = getattr(x, 'dtype', None)
            if _dtype is None: raise TypeError(f"Cannot find 'dtype' in examination isinstance({x}, {self}), please contact the developer for more information (Error Code: H123).")
            if len(self.base_type) > 1: raise TypeError(f"{x.__class__.__name__} object can only have one dtype, instead of {len(self.base_type)}: {self.base_type}")
            return issubclass(dtype(_dtype), self.base_type[0])
            # return self.base_type[0] == builtins.object or dtype == self.base_type[0] or self.base_type[0].__name__.split('.')[-1] in builtins.str(dtype)

        x = builtins.list(x)
        if self.base_size is not None:
            if len(self.base_size) != 1: raise TypeError(f"in {self}: array size should be a single integer of total number of elements")
            if self.base_size[0] >= 0 and len(x) != self.base_size[0]: return False
        if self.base_type is None: return True
        if len(self.base_type) > 1: raise TypeError(f"{x.__class__.__name__} object can only have one type, instead of {len(self.base_type)}: {self.base_type}")
        return all(isinstance(u, self.base_type[0]) for u in x)
    
    def __len__(self):
        if hasattr(self, 'element_types'): return len(self.element_types)
        if self.base_size is None: return -1
        p = 1
        for x in self.base_size:
            p *= x
            if x < 0: break
        else: return p
        return -1
    
    @property
    def lock_error(self):
        if self.base_type is None: return TypeError(f"Missing ending note '>>' after {get_type_name(self.origin_type)}<<.")
        base_type_str = ', '.join(get_type_name(t) for t in self.base_type)
        return TypeError(f"Missing ending note '>>' for element-wise type in {get_type_name(self.origin_type)}<<{base_type_str}.")
    def with_lock(self, locked): self.locked = locked; return self

def union(*types):
    if len(types) == 1 and isinstance(types[0], builtins.tuple): types = types[0]
    new_types = []
    for t in types:
        if getattr(t, 'op_type', None) == 'union': new_types.extend(t.types)
        elif isinstance(t, builtins.type): new_types.append(t)
        elif isinstance(t, builtins.tuple): new_types.extend(union(*t).types)
        elif isinstance(t, builtins.str): new_types.append(tag(t))
        elif t is None: new_types.append(builtins.object)
        else: as_type(t)
    return type('[' + ' | '.join(t.__name__ for t in new_types) + ']',
        op_type = 'union',
        types = new_types,
        __instancecheck__ = lambda _, x: isinstance(x, builtins.tuple(new_types)),
        __subclasscheck__ = lambda _, t: issubclass(t, builtins.tuple(new_types))
    )

@alias('intersect', 'insct')
def intersection(*types):
    if len(types) == 1 and isinstance(types[0], builtins.tuple): types = types[0]
    new_types = []
    for t in types:
        if getattr(t, 'op_type', None) == 'intersection': new_types.extend(t.types)
        elif isinstance(t, builtins.type): new_types.append(t)
        elif isinstance(t, builtins.tuple): new_types.append(union(*t)) # this is different to the line 'union' as it cannot be expanded here
        elif isinstance(t, builtins.str): new_types.append(tag(t))
        elif t is None: new_types.append(builtins.object)
        else: as_type(t)
    return type('[' + ' & '.join(t.__name__ for t in new_types) + ']',
        op_type = 'intersection',
        types = new_types,
        __instancecheck__ = lambda _, x: all(isinstance(x, mt) for mt in new_types),
        __subclasscheck__ = lambda _, t: all(issubclass(t, mt) for mt in new_types)
    )

def avoid(origin):
    if getattr(origin, 'op_type', None) == 'avoid': return origin.types
    elif getattr(origin, 'op_type', None) == 'union': return intersection(*[avoid(t) for t in origin.types])
    elif getattr(origin, 'op_type', None) == 'intersection': return union(*[avoid(t) for t in origin.types])
    return type(f"!{origin.__name__}",
        op_type = 'avoid', 
        types = origin,
        __instancecheck__ = lambda _, x: not isinstance(x, origin),
        __subclasscheck__ = lambda _, t: not issubclass(t, origin)
    )

def tag(string):
    return type(f"tag[{string}]",
        __subclasshook__ = lambda _, t: isinstance(t, builtins.type) and any(string in getattr(it, '__qualname__', builtins.str(it)).split('.') for it in t.__mro__)
    )

def note(string):
    return type(f"note[{string}]", __subclasshook__ = lambda _: True)

@alias("type_satisfies")
def class_satisfies(func):
    return type(f"check_class_by_[{func.__qualname__}]",
        __call__ = lambda _, x: func(x),
        __subclasshook__ = lambda _, x: func(x)
    )
    
def instance_satisfies(func):
    return type(f"check_instance_by_[{func.__qualname__}]",
        __call__ = lambda _, x: func(x),
        __instancecheck__ = lambda _, x: func(x)
    )

def __array_getitem__(self, index):
    """
    'array' is defined by subscripts ['type', element type such as 'dtype' and 'shape'].
        e.g. array[np.ndarray, np.int16, 3, 4] or array[torch.Tensor, 5, 8, 5]. 
    """
    if not isinstance(index, builtins.tuple): index = (index,)
    base_type = builtins.object
    base_dtype = None
    base_size = []
    stage = 0
    for arg in index:
        if stage == 0 and isinstance(arg, builtins.type) and not isinstance(arg, dtype):
            base_type = arg
            stage = 1
            continue
        elif stage == 0 and isinstance(arg, builtins.slice):
            if arg.step is not None: raise TypeError("Cannot identify 'array' type with two colons in subscript")
            base_type = arg.start
            base_dtype = dtype(arg.stop)
            stage = 2
            continue
        elif stage <= 1 and (isinstance(arg, dtype) or _touch(lambda: dtype(arg))):
            base_dtype = dtype(arg)
            stage = 2
            continue
        else: base_size.append(arg)
    base_size = builtins.tuple(base_size)
    if len(base_size) == 0: base_size = None
    elif len(base_size) == 1 and isinstance(base_size[0], builtins.tuple): base_size = base_size[0]
    size_str = '-' if base_size is None else repr(base_size).strip('(,)')
    base_type_str = f"<{base_dtype.__name__}>" if base_dtype is not None else ''
    return ArrayType(f"{get_type_name(base_type)}{base_type_str}[{size_str}]", base_type, base_type=base_dtype, base_size=base_size)

array = type("array",
    is_iterable = True,
    __getitem__ = __array_getitem__,
    __subclasshook__ = lambda _, t: isinstance(t, builtins.type) and hasattr(t, 'shape')
)

### >>> basic element objects <<< ###

object = type("object", builtins.object)
bool = type("bool", builtins.int, parent = builtins.bool)
int = type("int", builtins.int)
short = type("short", builtins.int, __dtype__ = 'int16')
long = type("long", builtins.int, __dtype__ = 'int64')
float = type("float", builtins.float)
half = type("half", builtins.int, __dtype__ = 'float16')
double = type("double", builtins.float, __dtype__ = 'float64')
rational = type("rational", float) # not carefully implemented yet
real = type("real", parent = union(int, float), op_type = 'union', types = [int, float])
complex = type("complex", builtins.complex)
scalar = type("scalar", parent = ArrayType('scalar', array, base_type=None, base_size=builtins.tuple()))
number = type("number", parent = union(scalar, real, complex, rational), op_type = 'union', types = [scalar, real, complex, rational])

property = type("property", builtins.property)
slice = type("slice", parent = builtins.slice)
null = type("null", parent = builtins.type(None))

### >>> functions and methods <<< ###

callable = instance_satisfies(builtins.callable)
functional = type("functional",
    __call__ = lambda _, x: not isinstance(x, builtins.type) and hasattr(x, '__call__'),
    __subclasshook__ = lambda _, t: isinstance(t, builtins.type) and not issubclass(t, builtins.type) and hasattr(t, '__call__')
)
lambda_ = lambda_func = type("lambda", parent = (lambda:...).__class__)
method = type("method", parent = __info__.__enter__.__class__)
function = type("function", parent = avoid.__str__.__class__)
builtin_function_or_method = type("builtin_function_or_method", parent = builtins.int.__new__.__class__)
method_descriptor = type("method_descriptor", parent = builtins.int.to_bytes.__class__)
method_wrapper = type("method_wrapper", parent = ''.__str__.__class__)
def __is_generator_func__(_, func):
    while hasattr(func, '__wrapped__'): func = func.__wrapped__
    return func.__code__.co_flags &0x20
generator_function = type("generator_function", functional, 
    __instancecheck__ = __is_generator_func__
)
classmethod = type("classmethod", builtins.classmethod)
staticmethod = type("staticmethod", builtins.staticmethod)

### >>> fixed element sequences <<< ###

str = type("str", builtins.str)
bytes = type("bytes", builtins.bytes)
bytearray = type("bytearray", builtins.bytearray)
memoryview = type("memoryview", parent = builtins.memoryview)
map = type("map", builtins.map)
filter = type("filter", builtins.filter)

### >>> array or sequences <<< ###

iterable = type("iterable", 
    is_iterable = True,
    __call__ = lambda _, x: hasattr(x, '__iter__') or hasattr(x, '__len__') or hasattr(x, '__getitem__'),
    __subclasshook__ = lambda _, t: isinstance(t, builtins.type) and (hasattr(t, '__iter__') or hasattr(t, '__len__') or hasattr(t, '__getitem__'))
)

sequence = type("sequence",
    is_iterable = True,
    __subclasshook__ = lambda _, t: isinstance(t, builtins.type) and (hasattr(t, '__iter__') and hasattr(t, '__len__'))
)

list = type("list", builtins.list, is_iterable = True)
tuple = type("tuple", builtins.tuple, is_iterable = True)
dict = type("dict", builtins.dict, is_iterable = True)
set = type("set", builtins.set, is_iterable = True)
reversed = type("reversed", builtins.reversed, is_iterable = True)
frozenset = type("frozenset", builtins.frozenset, is_iterable = True)

### >>> generators <<< ###

range = type("range", parent = builtins.range, is_iterable = True)
generator = type("generator", parent = (_ for _ in []).__class__, is_iterable = True)
zip = type("zip", builtins.zip, is_iterable = True)
enumerate = type("enumerate", builtins.enumerate, is_iterable = True)

if __name__ == "__main__":
    ...



try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "Types for typehint and overload. ",
    requires = ""
)

__all__ = """
    ArrayType
    ScalarType
    avoid
    union
    tag
    class_satisfies
    instance_satisfies
    
    object
    bool
    int
    long
    float
    double
    real
    complex
    scalar
    bool_scalar
    int_scalar
    long_scalar
    float_scalar
    double_scalar
    real_scalar
    complex_scalar
    property
    dtype
    type
    class_type
    super
    null
    
    callable
    functional
    lambda_func
    method
    function
    builtin_function_or_method
    method_descriptor
    method_wrapper
    classmethod
    staticmethod
    
    str
    bytearray
    bytes
    
    array
    iterable
    sequence
    list
    tuple
    dict
    set
    reversed
    frozenset
    
    zip
    enumerate
""".split()

import random, copy, builtins
from abc import ABCMeta

def _touch(f):
    try: return f()
    except Exception: ...

### >>> type generators <<< ###

class ArrayType(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, origin, args):
        if isinstance(args, (builtins.type, builtins.int)): args = (args,)

        get_name = lambda x: getattr(x, '__name__',
            f"{getattr(x.start, '__name__', x.start)}:{getattr(x.stop, '__name__', x.stop)}"
            if isinstance(x, builtins.slice) else builtins.str(x)
        ).split('.')[-1]
        if isinstance(args, builtins.list):
            arg_names = ', '.join(get_name(a) for a in args)
            type_name = f"{get_name(origin)}[{len(args)}]:[{arg_names}]"
        elif isinstance(args, builtins.tuple):
            type_names = ': '.join(get_name(t) for t in args if isinstance(t, builtins.type))
            size = builtins.tuple(s for s in args if not isinstance(s, builtins.type))
            if len(size) == 1: size = size[0]
            type_name = f"{get_name(origin)}[{type_names}, {size}]"

        if isinstance(args, builtins.list):
            return super().__new__(cls, type_name, (origin,), 
                builtins.dict(origin=origin, detail_mode='elements', element_types=args))
        if not isinstance(args, (builtins.tuple, builtins.type, builtins.int)):
            raise TypeError(f"Invalid ArrayType: {type_name}; ArrayType subscript only take list, tuple, type, or int. ")
        base_type = [builtins.object]
        base_size = -1
        if isinstance(args, builtins.tuple):
            base_type = []
            for arg in args:
                if isinstance(arg, builtins.type):
                    base_type.append(arg)
                elif isinstance(arg, builtins.tuple) and all(isinstance(a, builtins.type) or isinstance(a, dtype) for a in arg):
                    base_type.append(arg)
                elif isinstance(arg, builtins.slice):
                    if arg.step is not None:
                        raise TypeError(f"Invalid ArrayType: {type_name}; two colons in ArrayType subscript.")
                    base_type.append(arg.start)
                    base_type.append(arg.stop)
                else: break
            if len(base_type) <= 1: ...
            elif len(base_type) == 2 and issubclass(origin, builtins.dict): ...
            elif issubclass(origin, builtins.zip): ...
            else:
                raise TypeError(f"Invalid ArrayType: {type_name}; at most 2 types for dict key-value pair in subscript of ArrayType, one should use tuple of types to include multiple probabilities or use [(...),] if no size is given. ")
            base_size = args[len(base_type):]
            if len(base_size) == 1: base_size = base_size[0]
            if len(base_type) == 0: base_type = [builtins.object]
        return super().__new__(cls, type_name, (builtins.object,), 
            builtins.dict(origin=origin, detail_mode='cells', base_type=base_type, base_size=base_size))
    
    def __init__(self, origin, args): ...

    @property
    def __name__(self): return builtins.str(self)

    def __str__(self):
        get_name = lambda x: getattr(x, '__name__',
            f"{getattr(x.start, '__name__', x.start)}:{getattr(x.stop, '__name__', x.stop)}"
            if isinstance(x, builtins.slice) else builtins.str(x)
        ).split('.')[-1]
        if self.detail_mode == 'elements':
            arg_names = ', '.join(get_name(a) for a in self.element_types)
            return f"{get_name(self.origin)}[{len(self.element_types)}]:[{arg_names}]"
        elif self.detail_mode == 'cells':
            type_names = ': '.join(get_name(t) for t in self.base_type)
            return f"{get_name(self.origin)}[{type_names}, {self.base_size}]"
        else: raise RuntimeError(f"Unrecognized detail_mode {self.detail_mode}, please contact the developer for more information (Error Code: H122).")
        
    def __instancecheck__(self, x):
        if not isinstance(x, self.origin): return False
        if self.detail_mode == 'elements':
            y = builtins.list(x)
            if len(y) != len(self.element_types): return False
            return all(isinstance(u, t) for u, t in builtins.zip(y, self.element_types))
        elif self.detail_mode == 'cells':
            n_max_ele = 100
            if issubclass(self.origin, builtins.dict):
                if not isinstance(self.base_size, builtins.int): raise TypeError(f"in {self}: dict size should be an integer")
                if self.base_size >= 0 and len(x) != self.base_size: return False
                if len(self.base_type) == 1: self.base_type = [builtins.object, self.base_type[0]]
                if len(x) > n_max_ele: n_keys = n_max_ele
                else: n_keys = len(x)
                return all(isinstance(k, self.base_type[0]) and isinstance(x[k], self.base_type[1]) for k, _ in builtins.zip(x, builtins.range(n_max_ele)))
            if issubclass(self.origin, builtins.zip):
                if not isinstance(self.base_size, builtins.int): raise TypeError(f"in {self}: zip size should be an integer")
                y = copy.deepcopy(x)
                l = 0
                while True if self.base_size >= 0 else n_max_ele:
                    try: sample = next(y)
                    except StopIteration: break
                    if not all(isinstance(u, t) for u, t in builtins.zip(sample, self.base_type)): return False
                    l += 1
                if self.base_size >= 0 and self.base_size != l: return False
                return True
            if issubclass(self.origin, builtins.enumerate):
                if not isinstance(self.base_size, builtins.int): raise TypeError(f"in {self}: enumerate size should be an integer")
                y = copy.deepcopy(x)
                l = 0
                while True if self.base_size >= 0 else n_max_ele:
                    try: _, sample = next(y)
                    except StopIteration: break
                    if not isinstance(sample, self.base_type): return False
                    l += 1
                if self.base_size >= 0 and self.base_size != l: return False
                return True
            if hasattr(x, 'shape'):
                if self.base_size != -1:
                    if not isinstance(self.base_size, builtins.tuple): self.base_size = (self.base_size,)
                    if not (builtins.type(x.shape)(self.base_size) == x.shape or _touch(lambda: self.base_size == builtins.tuple(x.shape))): return False
                _dtype = getattr(x, 'dtype', None)
                if _dtype is None: raise TypeError(f"Cannot find 'dtype' in examination isinstance({x}, {self}), please contact the developer for more information (Error Code: H123).")
                if len(self.base_type) > 1: raise TypeError(f"{x.__class__.__name__} object can only have one dtype, instead of {len(self.base_type)}: {self.base_type}")
                return issubclass(dtype(_dtype), self.base_type[0])
                # return self.base_type[0] == builtins.object or dtype == self.base_type[0] or self.base_type[0].__name__.split('.')[-1] in builtins.str(dtype)
            if not isinstance(self.base_size, builtins.int): raise TypeError(f"in {self}: array size should be an integer")
            if len(self.base_type) > 1: raise TypeError(f"{x.__class__.__name__} object can only have one type, instead of {len(self.base_type)}: {self.base_type}")
            x = builtins.list(x)
            if self.base_size >= 0 and len(x) != self.base_size: return False
            return all(isinstance(u, self.base_type[0]) for u in x)
        else: raise RuntimeError(f"Unrecognized detail_mode {self.detail_mode}, please contact the developer for more information (Error Code: H122).")
    
    def __subclasscheck__(self, t):
        return issubclass(t, self.origin)
    
class ScalarType(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, *base_types):
        if len(base_types) == 1 and isinstance(base_types[0], builtins.tuple): base_types = base_types[0]
        name = f"Scalar[{', '.join(getattr(t, '__name__', builtins.str(t)) for t in base_types)}]"
        return super().__new__(cls, name, (builtins.object,), builtins.dict(base_types = base_types))
        
    def __init__(self, *base_types): ...

    @property
    def __name__(self): return builtins.str(self)

    def __str__(self): return f"Scalar[{', '.join(getattr(t, '__name__', builtins.str(t)) for t in self.base_types)}]"
    
    def __instancecheck__(self, x):
        if len(getattr(x, 'shape', [])) > 0: return False
        if len(self.base_types) == 0: return True
        for t in self.base_types:
            if isinstance(t, builtins.type):
                if isinstance(x, t): return True
            if isinstance(x, dtype(t)): return True
        return False

class avoid(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, origin):
        type_name = '!' + origin.__name__.split('.')[-1]
        return super().__new__(cls, type_name, (builtins.object,), builtins.dict(origin=origin))
    
    def __init__(self, origin): ...
    
    def __instancecheck__(self, x):
        return not isinstance(x, self.origin)
    def __subclasscheck__(self, t):
        return not issubclass(t, self.origin)
    
    def __str__(self): return self.__name__

class union(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, *types):
        types = tuple(t if isinstance(t, builtins.type) else tag(str(t)) for t in types)
        type_name = 'union[' + ', '.join(t.__name__.split('.')[-1] for t in types) + ']'
        return super().__new__(cls, type_name, (builtins.object,), builtins.dict(types=types))
    
    def __init__(self, *types): ...
    
    def __instancecheck__(self, x):
        return isinstance(x, self.types)
    def __subclasscheck__(self, t):
        return issubclass(t, self.types)
    
    def __str__(self): return self.__name__

class tag(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, text_tag):
        type_name = f"tag[{text_tag}]"
        return super().__new__(cls, type_name, (builtins.object,), builtins.dict(tag=text_tag))

    def __init__(self, text_tag): ...
    
    def __instancecheck__(self, x):
        return any(self.tag in getattr(t, '__qualname__', str(t)).split('.') for t in x.__class__.__mro__)
    def __subclasscheck__(self, t):
        return isinstance(t, builtins.type) and any(self.tag in getattr(it, '__qualname__', str(it)).split('.') for it in t.__mro__)
    
    def __str__(self): return self.__name__

class class_satisfies(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, is_subclass, is_instance=None):
        type_name = f"check_class_by_[{is_subclass.__name__}]"
        if is_instance is None: is_instance = lambda x: is_subclass(x.__class__)
        return super().__new__(cls, type_name, (builtins.object,), builtins.dict(subclasscheck=is_subclass, instancecheck=is_instance))

    def __init__(self, is_subclass, is_instance=None): ...
    def __call__(self, x): return self.mrois_subclass(x)
    
    def __instancecheck__(self, x):
        return self.instancecheck(x)
    def __subclasscheck__(self, t):
        return self.subclasscheck(t)
    
    def __str__(self): return self.__name__

class instance_satisfies(builtins.type):
    
    @builtins.staticmethod
    def __new__(cls, is_instance):
        type_name = f"check_instance_by_[{is_instance.__name__}]"
        return super().__new__(cls, type_name, (builtins.object,), builtins.dict(instancecheck=is_instance))

    def __init__(self, is_instance): ...
    def __call__(self, x): return self.instancecheck(x)
    
    def __instancecheck__(self, x):
        return self.instancecheck(x)
    def __subclasscheck__(self, t):
        return builtins.NotImplementedError
    
    def __str__(self): return self.__name__

### >>> basic element objects <<< ###

class object(builtins.object, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'object' is not regarded iterable")
        return ArrayType(iterable, (builtins.object, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.object in t.__mro__

class bool(builtins.int, metaclass=ABCMeta):
    
    @builtins.staticmethod
    def __new__(cls, arg=None):
        return super().__new__(cls, builtins.bool(arg))
    
    def __and__(self, y): return builtins.bool(self) and y
    def __rand__(self, x): return x and builtins.bool(self)
    def __or__(self, y): return builtins.bool(self) or y
    def __ror__(self, x): return x or builtins.bool(self)
    def __xor__(self, y): return builtins.bool(self).__xor__(y)
    def __rxor__(self, x): return builtins.bool(self).__rxor__(x)
    
    def __repr__(self): return builtins.bool(self).__repr__()
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'bool' is not iterable")
        return ArrayType(iterable, (builtins.bool, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.bool in t.__mro__

class int(builtins.int, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_and__(cls, other): print("in it!!")
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'int' is not iterable")
        return ArrayType(iterable, (builtins.int, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.int in t.__mro__

class long(builtins.int, metaclass=ABCMeta):
    
    __dtype__ = 'int64'
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'int' is not iterable")
        return ArrayType(iterable, (long, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.int in t.__mro__

class float(builtins.float, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'float' is not iterable")
        return ArrayType(iterable, (builtins.float, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.float in t.__mro__

class double(builtins.float, metaclass=ABCMeta):
    
    __dtype__ = 'float64'
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'float' is not iterable")
        return ArrayType(iterable, (double, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.float in t.__mro__

class real(builtins.property, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'real' is not iterable")
        return ArrayType(iterable, (real, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.int in t.__mro__ or builtins.float in t.__mro__

class complex(builtins.complex, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'complex' is not iterable")
        return ArrayType(iterable, (builtins.complex, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.complex in t.__mro__

scalar = ScalarType()
bool_scalar = ScalarType(bool)
int_scalar = ScalarType(int)
long_scalar = ScalarType(long)
float_scalar = ScalarType(float)
double_scalar = ScalarType(double)
real_scalar = ScalarType(real)
complex_scalar = ScalarType(complex)

class property(builtins.property, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'property' is not iterable")
        return ArrayType(iterable, (builtins.property, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.property in t.__mro__

class dtype(metaclass=ABCMeta):

    @builtins.staticmethod
    def __new__(cls, *args):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, builtins.str):
                module, *_, _dtype = arg.split('.')[0] if '.' in arg else [None, arg]
            elif isinstance(arg, builtins.type):
                module = getattr(arg, '__module__', None)
                _dtype = arg
                if any(getattr(t, '__module__', '').startswith('pyoverload') for t in arg.__mro__):
                    module = None
                    _dtype = getattr(arg, '__dtype__', arg)
                if not isinstance(_dtype, str) and any(getattr(t, '__module__', '').startswith('builtins') for t in arg.__mro__):
                    module = None
                    _dtype = arg.__name__.split('.')[-1]
            elif hasattr(arg, 'dtype') and isinstance(arg.dtype, dtype):
                module = getattr(arg.__class__, '__module__', None)
                _dtype = getattr(arg, 'dtype')
            elif isinstance(arg, dtype):
                module = getattr(arg.__class__, '__module__', None)
                _dtype = arg
            else: raise TypeError(f"Unrecognized dtype '{arg}' (of type {builtins.type(arg)})")

            class _this_dtype(builtins.type):

                def __getitem__(self, index):
                    class _sized_dtype(builtins.type):
                        def __instancecheck__(self, x):
                            if self.module is not None and self.module not in getattr(x.__class__, '__module__', ''): return False
                            if not hasattr(x, 'shape'): return False
                            if not (builtins.type(x.shape)(index) == x.shape or _touch(lambda: index == builtins.tuple(x.shape))): return False
                            if isinstance(self.dtype, builtins.type):
                                return isinstance(x[(0,) * len(x.shape)], self.dtype) or isinstance(x, self.dtype)
                            elif isinstance(self.dtype, dtype):
                                if not hasattr(x, 'dtype'): return False
                                return x.dtype == self.dtype
                            elif isinstance(self.dtype, builtins.str):
                                return self.dtype in builtins.str(getattr(x, 'dtype', ''))
                            else: raise TypeError(f"Invalid dtype of type {builtins.type(self.dtype)} read in dtype({self.dtype})[{builtins.str(index).strip('()')}]")
                        def __subclasscheck__(self, t): return self.module is None or self.module in getattr(t, 'module', getattr(t, '__module__', ''))
                    repr_dtype = repr(self.dtype).split('.')[-1]
                    return _sized_dtype(f"dtype[{repr_dtype}:{repr(index).strip('()')}]", (builtins.object,), builtins.dict(module = self.module, dtype = self.dtype))
                
                def __instancecheck__(self, x): return self[builtins.tuple()].__instancecheck__(x)
                def __subclasscheck__(self, t): return self.module is None or self.module in getattr(t, 'module', getattr(t, '__module__', ''))

            repr_dtype = repr(_dtype).split('.')[-1]
            return _this_dtype(f"dtype[{repr_dtype}]", (builtins.object,), builtins.dict(module = None if module is None else module.split('.')[0], dtype = _dtype))
        return super().__new__(cls, *args)
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'dtype' is not iterable")
        return ArrayType(iterable, (dtype, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t): return isinstance(t, builtins.type) and 'dtype' in t.__name__

class type(metaclass=ABCMeta):

    @builtins.staticmethod
    def __new__(cls, *args):
        if len(args) == 1: return args[0].__class__
        return super().__new__(cls, *args)
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'type' is not iterable")
        return ArrayType(iterable, (builtins.type, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.type in t.__mro__

class_type = type

class super(builtins.super, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'super' is not iterable")
        return ArrayType(iterable, (super, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return super in t.__mro__

class null(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'null' is not iterable")
        return ArrayType(iterable, (builtins.type(None), index))

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.type(None) in t.__mro__

### >>> functions and methods <<< ###

callable = instance_satisfies(builtins.callable)
# class callable(metaclass=ABCMeta):

#     @builtins.staticmethod
#     def __new__(cls, *args):
#         if len(args) == 1: return builtins.callable(args[0])
#         return super().__new__(cls, *args)
    
#     @builtins.classmethod
#     def __class_getitem__(cls, index):
#         if not isinstance(index, builtins.int):
#             raise TypeError("type 'callable' is not iterable")
#         return ArrayType(iterable, (callable, index))
    
#     @builtins.classmethod
#     def __subclasshook__(cls, t):
#         return isinstance(t, builtins.type) and hasattr(t, '__call__')

class functional(metaclass=ABCMeta):

    @builtins.staticmethod
    def __new__(cls, *args):
        if len(args) == 1: return isinstance(args[0], functional)
        return super().__new__(cls, *args)
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'functional' is not iterable")
        return ArrayType(iterable, (functional, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return isinstance(t, builtins.type) and not issubclass(t, builtins.type) and hasattr(t, '__call__')

class lambda_func(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'function' is not iterable")
        return ArrayType(iterable, ((lambda:...).__class__, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return (lambda:...) in t.__mro__
    
    def __call__(self, *args, **kwargs): ...

class method(metaclass=ABCMeta):

    @builtins.staticmethod
    def __new__(cls, *args):
        if len(args) == 2: return __info__.__enter__.__class__(*args)
        return super().__new__(cls, *args)
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'method' is not iterable")
        return ArrayType(iterable, (__info__.__enter__.__class__, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return __info__.__enter__.__class__ in t.__mro__

class function(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'function' is not iterable")
        return ArrayType(iterable, (avoid.__str__.__class__, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return avoid.__str__.__class__ in t.__mro__
    
    def __call__(self, *args, **kwargs): ...

class builtin_function_or_method(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'builtin_function_or_method' is not iterable")
        return ArrayType(iterable, (builtins.int.__new__.__class__, index))

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.int.__new__.__class__ in t.__mro__

class method_descriptor(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'method_descriptor' is not iterable")
        return ArrayType(iterable, (builtins.int.to_bytes.__class__, index))

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.int.to_bytes.__class__ in t.__mro__

class method_wrapper(metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'method_wrapper' is not iterable")
        return ArrayType(iterable, (''.__str__.__class__, index))

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return ''.__str__.__class__ in t.__mro__

class classmethod(builtins.classmethod, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'classmethod' is not iterable")
        return ArrayType(iterable, (builtins.classmethod, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.classmethod in t.__mro__

class staticmethod(builtins.staticmethod, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("type 'staticmethod' is not iterable")
        return ArrayType(iterable, (builtins.staticmethod, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.staticmethod in t.__mro__

### >>> fixed element sequences <<< ###

class str(builtins.str, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("cannot define the type of elements in 'str'")
        return ArrayType(iterable, (builtins.str, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.str in t.__mro__

class bytes(builtins.bytes, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("cannot define the type of elements in 'bytes'")
        return ArrayType(iterable, (builtins.int, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.bytearray in t.__mro__

class bytearray(builtins.bytearray, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        if not isinstance(index, builtins.int):
            raise TypeError("cannot define the type of elements in 'bytearray'")
        return ArrayType(iterable, (builtins.int, index))
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.bytearray in t.__mro__

### >>> array or sequences <<< ###

class array(metaclass=ABCMeta):

    @builtins.classmethod
    def __class_getitem__(cls, index):
        """
        'array' is defined by subscripts ['type', element type such as 'dtype' and 'shape'].
            e.g. array[np.ndarray, np.int16, 3, 4] or array[torch.Tensor, 5, 8, 5]. 
        """
        if not isinstance(index, builtins.tuple): index = (index,)
        base_type = builtins.object
        base_dtype = scalar
        base_size = []
        stage = 0
        for arg in index:
            if stage == 0 and isinstance(arg, builtins.type) and not isinstance(arg, dtype):
                base_type = arg
                stage = 1
                continue
            elif stage == 0 and isinstance(arg, builtins.slice):
                if arg.step is not None: raise TypeError("Cannot identify 'array' type with two colons in subscript")
                base_type = arg.start
                base_dtype = dtype(arg.stop)
                stage = 2
                continue
            elif stage <= 1 and (isinstance(arg, dtype) or _touch(lambda: dtype(arg))):
                base_dtype = dtype(arg)
                stage = 2
                continue
            else: base_size.append(arg)
        if len(base_size) == 0: base_size = [-1]
        return ArrayType(base_type, (base_dtype, *base_size))

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return isinstance(t, builtins.type) and hasattr(t, 'shape') and hasattr(t, 'dtype')

class iterable(metaclass=ABCMeta):

    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(iterable, index)

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return isinstance(t, builtins.type) and (hasattr(t, '__iter__') or hasattr(t, '__len__') or hasattr(t, '__getitem__'))

class sequence(metaclass=ABCMeta):

    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(sequence, index)

    @builtins.classmethod
    def __subclasshook__(cls, t):
        return isinstance(t, builtins.type) and (hasattr(t, '__iter__') and hasattr(t, '__len__'))

class list(builtins.list, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __not__(cls): return not_list
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.list, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.list in t.__mro__

class tuple(builtins.tuple, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.tuple, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.tuple in t.__mro__

class dict(builtins.dict, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.dict, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.dict in t.__mro__

class set(builtins.set, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.set, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.set in t.__mro__

class reversed(builtins.reversed, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.reversed, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.reversed in t.__mro__

class frozenset(builtins.frozenset, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.frozenset, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.frozenset in t.__mro__

### >>> generators <<< ###

class zip(builtins.zip, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.zip, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.zip in t.__mro__

class enumerate(builtins.enumerate, metaclass=ABCMeta):
    
    @builtins.classmethod
    def __class_getitem__(cls, index):
        return ArrayType(builtins.enumerate, index)
    
    @builtins.classmethod
    def __subclasshook__(cls, t):
        return builtins.enumerate in t.__mro__

if __name__ == "__main__":
    ...



from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "Uncategorized features, not included in the main package.",
    help = "Use `pycamia.more` to call the functions. "
)

__all__ = """
    once
    void
    random_seed
""".split()

import sys

class Void: ...
void = Void()

class OnceError(Exception):
    pass

class Once:

    def __init__(self):
        self.history = set()
        self._disabled = False

    def filename_and_linenu(self):
        try:
            raise Exception
        except:
            f = sys.exc_info()[2].tb_frame.f_back.f_back
        return (f.f_code.co_filename, f.f_lineno)

    def enable(self):
        self._disabled = False

    def disable(self):
        self._disabled = True

    def __bool__(self):
        if self._disabled:
            return False
        f_name, l_nu = self.filename_and_linenu()
        if (f_name, l_nu) in self.history:
            return False
        self.history.add((f_name, l_nu))
        return True

once = Once()

def random_seed(seed):
    import random
    import numpy
    import batorch
    import torch
    random.seed(seed)
    numpy.random.seed(seed)
    batorch.manual_seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic=True
    torch.backends.cudnn.benchmark = True




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2022-02",
    fileinfo = "File of important decorators.",
    help = "Use `@decorator` before functions. "
)

__all__ = """
    alias
    decorator
    restore_type_wrapper
""".split()

from functools import wraps
from .environment import get_environ_vars

def decorator(wrapper_func):
    """
    A decorator decorating decorators, i.e. decorating a function decoration 
        `dec` to enable calling it by `@dec` or `@dec(*args, **kwargs)`.
    Note 1: The first argument of the wrapped decorator `dec` should be `func=None`. 
    Note 2: The return value for `dec(*args, **kwargs)` should be a decorator.
    Note 3: Decorator `@wraps` is automatically added for wrappers in coding. 

    Examples::
    >>> @decorator
    ... def function_wrapper(func=None, activated=True):
    ...     def wrapper(*args, **kwargs):
    ...         print("activated")
    ...         return func(*args, **kwargs)
    ...     if func is None: return function_wrapper if activated else (lamba f: f)
    ...     return wrapper
    ... 
    >>> @function_wrapper
    ... def func1(x): return x + 1
    ...
    >>> @function_wrapper(activated=True)
    ... def func2(x): return x + 1
    ...
    >>> @function_wrapper(activated=False)
    ... def func3(x): return x + 1
    ...
    >>> print(func1(1), func2(2), func3(3))
    activated (1,)
    activated (2,)
    2 3 4
    """
    if not callable(wrapper_func):
        raise RuntimeError(f"@decorator wrapping a non-wrapper: {wrapper_func}")
    def wrapper(*args, **kwargs):
        if (len(args) == 1 and callable(args[0]) or # functions
            len(args) == 2 and callable(args[1])): # methods
            func = args[-1]
            raw_func = func.__func__ if isinstance(func, __info__.__enter__.__class__) else func
            raw_func = getattr(raw_func, '__wrapped__', raw_func)
            func_name = f"{raw_func.__name__}:{wrapper_func.__name__}"
            outer_func = wraps(raw_func)(wrapper_func(*args, **kwargs))
            outer_func.__name__ = func_name
            outer_func.__doc__ = raw_func.__doc__
            if isinstance(func, staticmethod): trans = staticmethod
            elif isinstance(func, classmethod): trans = classmethod
            else: trans = lambda x: x
            return trans(outer_func)
        try: return decorator(wrapper_func(*args, **kwargs))
        except RuntimeError as e:
            print(f"Error when decorating by {wrapper_func.__qualname__}")
            raise e
        # raise TypeError(f"Invalid decorator '@{wrapper_func.__qualname__}' for {args}")
    return wraps(wrapper_func)(wrapper)

class alias: 
    """
    Create an alias function of the decorated function. 

    Note: 
        One can also use it to return the value after assignment in python<3.9.

    Examples::
        >>> @other_wrappers # wrappers for function `func_b` only. 
        ... @alias("func_a", b=1) # wrappers between two aliases are for functions `func_a` and `func_b`.
        ... @alias("func_c", b=2)
        ... @some_wrappers # wrappers for functions `func_a`, `func_b` and `func_c`.
        ... def func_b(x, b):
        ...     print(x+b)
        ... 
        >>> func_a(1), func_b(2, 4), func_c(7)
        (2, 6, 9)
        >>> def compose(f, g):
        ...     return lambda x: f(g(x))
        >>> compose(alias("square")(lambda x: x**2), square)(3)
        81
    
    Examples for Extended Usages:
        >>> a = alias('b')(10)
        >>> a, b
        (10, 10)
        >>> class A:
        ...     @alias('a')
        ...     @property
        ...     def name(self): return 1
        ...     @alias('a')
        ...     @name.setter
        ...     def name(self, value): pass
        ...
        >>> A().a, A().name
        (1, 1)
        >>> A().a = 2
    """
    def __init__(self, *names, **kwargs): self.names = names; self.kwargs = kwargs
    def __call__(self, func):
        vars = get_environ_vars()
        if len(self.kwargs) > 0:
            @wraps(func)
            def wrapper(*args, **kwargs):
                kwargs.update(self.kwargs)
                return func(*args, **kwargs)
        else: wrapper = func
        for n in self.names:
            vars[n] = wrapper
        return func

def _restore_type_wrapper(func, special_attr):
    @wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        if len(args) == 0: return ret
        if str(type(args[0])) in func.__qualname__ and len(args) > 1: totype = type(args[1])
        else: totype = type(args[0])
        constructor = totype
        if "numpy.ndarray" in str(totype):
            import numpy as np
            constructor = np.array
        elif "batorch" in str(totype):
            import batorch as bt
            constructor = lambda x: x.as_subclass(bt.Tensor) if isinstance(x, torch.Tensor) else bt.tensor(x)
        elif "torch.Tensor" in str(totype):
            import torch
            constructor = lambda x: x.as_subclass(torch.Tensor) if isinstance(x, torch.Tensor) else torch.tensor(x)
        if not isinstance(ret, tuple): ret = (ret,)
        output = tuple()
        for r in ret:
            try: new_r = constructor(r)
            except: new_r = r
            for a in special_attr:
                if a in dir(r): exec(f"new_r.{a} = r.{a}")
            output += (new_r,)
        if len(output) == 1: output = output[0]
        return output
    return wrapper

def restore_type_wrapper(*args):
    """
    Restore type for array/tensor objects, as the first input. 
    """
    if len(args) == 1 and callable(args[0]):
        return _restore_type_wrapper(args[0], [])
    else:
        def restore_type_decorator(func):
            return _restore_type_wrapper(func, args)
        return restore_type_decorator





from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2022-02",
    fileinfo = "File to format transformation files from `zxhtools`.",
    requires = ["numpy", "batorch", "pycamia"]
).check()

__all__ = """
    AFF FFD DDF FLD TRS
    transform
    zxhtransform
""".split()

from ..trans import Affine, DenseDisplacementField, FreeFormDeformation
from ..trans import ComposedTransformation, Quaterns2Matrix, interpolation
from ..stdio import IMG

with __info__:
    import numpy as np
    import batorch as bt
    from pycamia import SPrint, alias, Path

class TRS:

    def __init__(self, **params):
        self.__dict__.update(params)
        self.path = None

    @classmethod
    def load(cls, p):
        p = Path(p)
        with p.open() as fp:
            self = cls.parse(fp.read())
            self.path = Path(p)
            return self

    @classmethod
    def parse(cls, string):
        blocks = string.strip('\n#').split('\n#')
        variables = [tuple(x.strip() for x in b.split('#')) for b in blocks]
        var_names = [x[0].replace(' ', '_').replace(',', '__c__') for x in variables]
        variables = dict(variables)
        params = {}
        for var in variables:
            value = variables[var]
            try:
                value = [[[[eval(d.strip()) for d in c.strip().split(' ')] for c in b.strip().split('\t')] for b in a.strip().split('\n')] for a in value.strip().split('\n\n')]
                value = np.array(value).squeeze()
            except (NameError, SyntaxError): pass
            if isinstance(value, str):
                value = value.replace('\\', '/')
            params[var.replace(' ', '_').replace(',', '__c__')] = value
        if cls == TRS:
            if any(['matri' in k for k in params.keys()]): cls = AFF
            elif any(['control_point' in k for k in params.keys()]): cls = FFD
            elif any(['field' in k for k in params.keys()]): cls = FLD
        return cls(**params, var_names=var_names)

    def __str__(self):
        output = SPrint()
        for key in self.var_names:
            name = key.replace('__c__', ',').replace('_', ' ')
            output("#{}#".format(name))
            value = getattr(self, key)
            if isinstance(value, str): output(value)
            if value.ndim == 0: output(value)
            if value.ndim == 1: output(*[self.pattern.get(name, {}).get('format', '%d')%v for v in value], sep=self.pattern.get(name, {}).get('sep', '\n'))
            if value.ndim == 2: output(*['\t'.join([self.pattern.get(name, {}).get('format', '%f')%v for v in c]) for c in value.tolist()], sep='\n')
            if value.ndim > 2: output(*['\n'.join(['\t'.join([' '.join([self.pattern.get(name, {}).get('format', '%f')%v for v in c]) for c in b]) for b in a]) for a in value.tolist()], sep='\n\n')
        return output.text

    def save(self, p):
        p = Path(p)
        with p.open('w') as fp:
            fp.write(str(self))

    def get(self, key):
        return getattr(self, key.replace(' ', '_').replace(',', '__c__'))
    
    def to_image_space(self, source, target):
        if isinstance(source, str): source = IMG(source).affine
        if isinstance(target, str): target = IMG(target).affine
        return self.trans.to('image', source_affine=source, target_affine=target)
    
class AFF(TRS):

    pattern = {
        'matrix': {'format': '%.6f'},
        'skew in x-y and y-x plane': {'format': '%.6f'},
        'skew in y-z and z-y plane': {'format': '%.6f'},
        'skew in z-x and x-z plane': {'format': '%.6f'},
        'rotation center': {'sep': '\t', 'format': '%.6f'},
        'rotation angle': {'sep': '\t', 'format': '%.6f'},
        'scaling': {'sep': '\t', 'format': '%.6f'},
        'translation': {'sep': '\t', 'format': '%.6f'}
    }
    
    @property
    def trans(self):
        if not hasattr(self, '_trans'):
            self._trans = Affine(bt.tensor(self.matrix).float())
        return self._trans

class FFD(TRS):

    pattern = {
        'control grid spacing': {'format': '%15.6f'},
        'offsets of the control points': {'format': '%.6f'},
        'quaternion qf,qb,qc,qd,qx,qy,qz': {'sep': '\t', 'format': '%.6f'}
    }
    
    @property
    def trans(self):
        if not hasattr(self, '_trans'):
            qf,qb,qc,qd,qx,qy,qz = self.get('quaternion qf,qb,qc,qd,qx,qy,qz')
            affmat = Quaterns2Matrix([qb,qc,qd,qx,qy,qz]).float()
            if qf == -1: affmat = affmat @ bt.diag([1., 1., -1., 1.])
            offsets = bt.tensor(self.offsets_of_the_control_points).float().T.channel_dimension_(0)
            # grid_spacing = (bt.inv(affmat)[0, :3, :3] @ bt.tensor(self.control_grid_spacing).float().unsqueeze(-1)).abs().squeeze().tolist()
            ffd = FreeFormDeformation(offsets, spacing = self.control_grid_spacing)
            self._trans = ComposedTransformation(Affine(affmat), ffd, Affine(bt.inv(affmat)), mode="spatial")
        return self._trans

@alias('DDF')
class FLD(TRS):

    pattern = {}
    
    @property
    def trans(self):
        if not hasattr(self, '_trans'):
            pathx = Path(self.get('transformation fields in dimension x'))
            pathy = Path(self.get('transformation fields in dimension y'))
            pathz = Path(self.get('transformation fields in dimension z'))
            if self.path is not None:
                pathx = self.path.parent/pathx.filename
                pathy = self.path.parent/pathy.filename
                pathz = self.path.parent/pathz.filename
            dx = IMG(pathx).to_tensor()
            dy = IMG(pathy).to_tensor()
            dz = IMG(pathz).to_tensor()
            disp = bt.stack(dx, dy, dz, {})
            self._trans = DenseDisplacementField(disp)
        return self._trans
    
    @property
    def space(self):
        if not hasattr(self, '_disp'):
            pathx = Path(self.get('transformation fields in dimension x'))
            pathy = Path(self.get('transformation fields in dimension y'))
            pathz = Path(self.get('transformation fields in dimension z'))
            if self.path is not None:
                pathx = self.path.parent/pathx.filename
                pathy = self.path.parent/pathy.filename
                pathz = self.path.parent/pathz.filename
            disp = bt.stack(IMG(pathx).to_tensor(), IMG(pathy).to_tensor(), IMG(pathz).to_tensor(), {})
            self._disp = disp
        return self._disp
    
def transform(target, source, output=None, *trans, **kwargs):
    if output is None: output = Path(source)^target.name.split('_')[-1]
    target_img = IMG(target)
    all_trans = []
    for t in trans:
        x = TRS.load(t)
        if isinstance(x, FLD):
            if 'SIC' in str(t) or 'SIC' in x.get('transformation fields in dimension x'):
                sp = x.space
                all_trans.append(DenseDisplacementField(sp - bt.image_grid(sp.space)).to_world_space(source, target))
                continue
        all_trans.append(x.trans)
    target_img.save(interpolation(IMG(source), ComposedTransformation(*all_trans).to_image_space(source, target), target_space=target_img.shape, **kwargs), output)
    
def zxhtransform(target, source, *trans):
    return transform(target, source, *trans[::-1])


if __name__ == '__main__':
    aff = AFF.load('image2.AFF')
    ffd = FFD.load('image2.FFD')
    print(aff, ffd)




__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to manage the environment.",
    help = "Use `get_**` to obtain the the variables etc. outside the function. "
)

__all__ = """
    get_environ_vars
    get_environ_globals
    get_environ_locals
    update_locals_by_environ
    get_args_expression
    get_reference_line
    get_declaration
    EnvironVars
""".split()

import os, re, sys, builtins
from .strop import tokenize
from .pythonop import python_line

try:
    import ctypes
except ModuleNotFoundError:
    ctypes = None
else:
    if hasattr(ctypes, "pythonapi") and \
       hasattr(ctypes.pythonapi, "PyFrame_LocalsToFast"): pass
    else: ctypes = None

stack_error = lambda x, ext: TypeError(f"Unexpected function stack for {x}, please contact the developer for further information (Error Code: E001). {ext}")

def _get_frames(i = [2, 3], key=''):
    """
    Get frames in stack. 
    By default: it gets frame of the function calling get_environ (function frame) and the frame calling this function (client frame). 
    i = -1 for all frames
    Returns: function frame, client frame
    """
    frames = []
    frame = sys._getframe()
    fname = frame.f_back
    if isinstance(i, int): i = [i]
    if i is not None:
        if len(i) == 0: raise IndexError("Invalid index for _get_frames")
        max_i = max(i)
    while frame is not None:
        frame_file = frame.f_code.co_filename
        # if frame_file.startswith('<') and frame_file.endswith('>') and frame_file != '<stdin>':
        #     frame = frame.f_back
        #     continue
        if i is None:
            if frame.f_code.co_name == key: frames.append(frame)
        elif i == [-1]:
            if frame is None: return frames
            frames.append(frame) 
        else:
            frames.append(frame)
            if len(frames) >= max_i + 1:
                domain = [frames[j] for j in i]
                if key != '': domain = [f for f in domain if f.f_code.co_name == key]
                return domain if len(domain) > 1 else domain[0]
        frame = frame.f_back
    if i is not None and i != [-1] or len(frames) == 0:
        try: f_all = _get_frames(-1)
        except: raise stack_error(fname, f"\n_get_frames({i}) got stack: \n" + '\n'.join(map(str, frames)))
        raise stack_error(fname, "\nComplete stack: \n" + '\n'.join(map(str, f_all)) + f"\n_get_frames({i}) got stack: \n" + '\n'.join(map(str, frames)))
    return frames

def _update_locals(frame):
    if ctypes is not None:
        ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))

class VarDict(dict):

    def __new__(cls, frame, method = 'global'):
        self = super().__new__(cls, getattr(frame, f"f_{method}s"))
        self.frame = frame
        self.method = method
        return self
    
    def __init__(self, frame, method):
        super().__init__(getattr(frame, f"f_{method}s"))
        
    @property
    def data(self): return getattr(self.frame, f"f_{self.method}s")

    def set(self, name, value):
        self[name] = value

    def __getitem__(self, name):
        if name not in self.data:
            if hasattr(builtins, name): return getattr(builtins, name)
            raise NameError(f"No variable '{name}' found in the {self.method} environment. ")
        return self.data[name]

    def __setitem__(self, name, value):
        super().__setitem__(name, value)
        if self.method == 'local':
            self.frame.f_locals[name] = value
            _update_locals(self.frame)
        else: self.frame.f_globals[name] = value

    def __getattr__(self, key):
        if key in ['frame', 'method', 'set'] + dir(dict): return super().__getattribute__(key)
        return self[key]

    def __setattr__(self, key, value):
        if key in ['frame', 'method', 'set'] + dir(dict): return super().__setattr__(key, value)
        self[key] = value

class EnvironVars:
    
    def __init__(self, frame): self.frame = frame

    def get(self, name, default=None):
        return self.locals.get(name, self.globals.get(name, default))

    def set(self, name, value, in_dict=None):
        if not in_dict: in_dict = 'local'
        if len(self.locals) == len(self.globals): in_dict = 'global'
        if in_dict.lower().startswith('loc'): self.locals.set(name, value)
        else: self.globals.set(name, value)
    
    def update(self, dic, in_dict=None):
        for k, v in dic.items(): self.set(k, v, in_dict=in_dict)
        
    def __contains__(self, key): return key in self.locals or key in self.globals
    def __getitem__(self, key):
        if key in self.locals: return self.locals[key]
        return self.globals[key]
    def __setitem__(self, key, value): self.set(key, value)
    def __getattr__(self, key):
        if key in ['frame', 'locals', 'globals']: return super().__getattr__(key)
        return self[key]
    def __setattr__(self, key, value):
        if key in ['frame', 'locals', 'globals']: return super().__setattr__(key, value)
        return self.set(key, value)
    
    @property
    def locals(self): return VarDict(self.frame, method='local')
    
    @property
    def globals(self): return VarDict(self.frame, method='global')
    
    @property
    def all(self):
        all_dict = self.frame.f_globals.copy()
        all_dict.update(self.frame.f_locals)
        return VarDict(self.frame, all_dict)
    
    def __str__(self):
        return "[ locals]: " + str(self.locals) + '\n[globals]: ' + str(self.globals)

def get_environ_vars(offset=0, pivot=''):
    """If there is a function 'f' called in script 's', one can use 'get_environ_vars' in 'f' to obtain the variables in 's'.
    Args:
        offset (int, optional): If 's' calls 'u' and 'u' calls 'f', one need to set 'offset' to 1 for the additional scope 'u'. Defaults to 0.
        pivot (str, optional): If one has the name of the function of which the variables are needed, place it here. The most previous call would be obtained. Defaults to ''.
    """
    client_frame = _get_frames(3) # offset of frame
    if pivot: client_frame = _get_frames(None, key=pivot)
    if isinstance(client_frame, list): client_frame = client_frame[-1]
    for _ in range(offset):
        client_frame = client_frame.f_back
    return EnvironVars(client_frame)

def get_environ_globals(offset=0, pivot=''):
    """If there is a function 'f' called in script 's', one can use 'get_environ_globals' in 'f' to obtain the global variables in 's'.
    Args:
        offset (int, optional): If 's' calls 'u' and 'u' calls 'f', one need to set 'offset' to 1 for the additional scope 'u'. Defaults to 0.
        pivot (str, optional): If one has the name of the function of which the global variables are needed, place it here. The most previous call would be obtained. Defaults to ''.
    """
    return get_environ_vars(offset, pivot).globals

def get_environ_locals(offset=0, pivot=''):
    """If there is a function 'f' called in script 's', one can use 'get_environ_locals' in 'f' to obtain the local variables in 's'.

    Note: Changing the values in the local object will not be assigned to the original scope, please use methods 'set' or 'update' of variables object obtained by 'get_environ_vars' to perform assignment.

    Args:
        offset (int, optional): If 's' calls 'u' and 'u' calls 'f', one need to set 'offset' to 1 for the additional scope 'u'. Defaults to 0.
        pivot (str, optional): If one has the name of the function of which the local variables are needed, place it here. The most previous call would be obtained. Defaults to ''.
    """
    return get_environ_vars(offset, pivot).locals

def update_locals_by_environ():
    module_frame, client_frame = _get_frames()
    vars_set = client_frame.f_locals.copy()
    vars_set.update(module_frame.f_locals)
    module_frame.f_locals.update(vars_set)
    _update_locals(module_frame)

class StrIO:
    def __init__(self, file_name = os.path.abspath('.null')):
        self._str_ = None
        self._file_ = open(file_name, 'w+')
        self.file_name = file_name
        self.fileno = self._file_.fileno
    def write(self, s): self._file_.write(s)
    def __str__(self):
        if self._str_ is not None: return self._str_
        self._file_.seek(0)
        self._str_ = self._file_.read()
        self.close()
        return self._str_
    def split(self, c=None): return str(self).split(c)
    def string(self): return str(self)
    def close(self):
        self._file_.close()
        if self.file_name == os.path.abspath('.null'):
            os.remove(self.file_name)
            self._file_ = None
    
def get_args_expression(func_name = None, search_more=False, with_line_info=False):
    _, _, module_frame, *client_frames = _get_frames(i=-1)
    if func_name is None: func_name = module_frame.f_code.co_name
    if func_name == '<lambda>':
        if os.path.exists(module_frame.f_code.co_filename):
            with open(module_frame.f_code.co_filename) as fp:
                for _ in range(module_frame.f_lineno-1): fp.readline()
                l = python_line(fp)
                func_name = l.split("lambda", 1)[0].split(';')[-1].rstrip(" =").strip()
    for client_frame in client_frames:
        if os.path.exists(client_frame.f_code.co_filename):
            with open(client_frame.f_code.co_filename) as fp:
                for _ in range(client_frame.f_lineno-1): fp.readline()
                l = python_line(fp)
                if func_name not in l:
                    if search_more: continue
                    raise TypeError(f"Cannot find function name `{func_name}` in {client_frame.f_code.co_filename} line {client_frame.f_lineno}:\n\t> {l.lstrip()}\n" +
                        "Problem occurs in code stack, please contact the developer for further information (Error Code: E002). ")
                error = TypeError(f"Cannot get args expression. Please avoid using two/none function '{func_name}'s in a single code line: {l}. ")
                lines_with_func = [x for x in tokenize(l, sep=[';']) if func_name in x]
                if len(lines_with_func) > 1: raise error
                line = lines_with_func[0]
                parts = re.split(rf'[^\w]{func_name}([^\w])', ' ' + line)
                if len(parts) != 3: raise error
                exp = tokenize(''.join(parts[1:]), sep=['.', ')'])[0]
                while True:
                    if exp.startswith('(') and exp.endswith(')'): exp = exp[1:-1]
                    else: break
                if with_line_info: return client_frame.f_code.co_filename, client_frame.f_lineno, exp
                else: return exp
        elif with_line_info: return None, None, "<unreachable arg expression>"
        else: return "<unreachable arg expression>"
    raise TypeError(f"Cannot find function name `{func_name}` in {client_frame.f_code.co_filename} line {client_frame.f_lineno}:\n\t{l}" +
        "Problem occurs in code stack, please contact the developer for further information (Error Code: E002). ")
    
def get_reference_line(func_name = None, search_more=False, with_line_info=False):
    _, _, module_frame, *client_frames = _get_frames(i=-1)
    if func_name is None: func_name = module_frame.f_code.co_name
    for client_frame in client_frames:
        if os.path.exists(client_frame.f_code.co_filename):
            with open(client_frame.f_code.co_filename) as fp:
                for _ in range(client_frame.f_lineno-1): fp.readline()
                l = python_line(fp)
                if func_name not in l:
                    if search_more: continue
                    raise TypeError(f"Cannot find function name `{func_name}` in {client_frame.f_code.co_filename} line {client_frame.f_lineno}:\n\t{l}" +
                        "Problem occurs in code stack, please contact the developer for further information (Error Code: E002). ")
                if with_line_info: return client_frame.f_code.co_filename, client_frame.f_lineno, l
                else: return l
        elif with_line_info: return None, None, "<unreachable reference>"
        else: return "<unreachable reference>"
    raise TypeError(f"Cannot find function name `{func_name}` in {client_frame.f_code.co_filename} line {client_frame.f_lineno}:\n\t{l}" +
        "Problem occurs in code stack, please contact the developer for further information (Error Code: E002). ")

def get_declaration(func, func_name = None):
    func_code = func.__code__
    if func_name is None: func_name = func.__name__
    if os.path.exists(func_code.co_filename):
        with open(func_code.co_filename) as fp:
            for _ in range(func_code.co_firstlineno - 1): fp.readline()
            i = func_code.co_firstlineno
            while True:
                l = python_line(fp)
                if func_name not in l:
                    if not l.strip() or l.strip().startswith('@') or l.strip().startswith('#'): i += 1; continue
                    raise TypeError(f"Cannot find function name `{func_name}` in {func_code.co_filename} line {i}:\n\t{l}" +
                        "Problem occurs in code stack, please contact the developer for further information (Error Code: E003). ")
                dec_line = l
                break
    else:
        ss = StrIO()
        oldout = sys.stdout
        sys.stdout = ss
        help(func)
        sys.stdout = oldout
        dec_line = [l for l in python_lines(ss) if len(l) > 0 and 'Help' not in l][0]
    if not dec_line: return "<unreachable declaration>"
    dec = dec_line.strip().split(' ', 1)[-1].rstrip(':')
    return dec




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of list operations. "
)

__all__ = """
    prod
    cumsum
    cumprod
    cartesian_prod
    argmin
    argmax
    min_argmin
    max_argmax
    kth_biggest
    kth_smallest
    median
    flatten_list
    item
    to_list
    to_tuple
    to_set
    map_ele
    sublist
    arg_tuple
    arg_extract
    count
    unique
    infinite_itemize
    cat_generator
    param_join
""".split()

from .exception import avouch, touch
from typing import Iterable
from random import randint

def prod(x, init=1):
    """
    Returns the product of elements, just like built-in function `sum`.
    
    Examples::
        >>> prod([5, 2, 1, 4, 2])
        80
    """
    for i in x: 
        if hasattr(i, "__mul__") or hasattr(i, "__rmul__"):
            init *= i
    return init

def cumsum(x, init=0):
    """
    Returns the cumulated sum of elements.
    
    Examples::
        >>> cumsum([5, 2, 1, 4, 2], 0)
        [5, 7, 8, 12, 14]
    """
    res = [init]
    for i in x: 
        if hasattr(i, "__add__") or hasattr(i, "__radd__"):
            res.append(res[-1] + i)
    return res[1:]

def cumprod(x, init=1):
    """
    Returns the cumulated product of elements.
    
    Examples::
        >>> cumsum([5, 2, 1, 4, 2], 2)
        [10, 20, 20, 80, 160]
    """
    res = [init]
    for i in x: 
        if hasattr(i, "__mul__") or hasattr(i, "__rmul__"):
            res.append(res[-1] * i)
    return res[1:]

def cartesian_prod(x, *y):
    """
    Returns the cartesian product of lists.
    
    Examples::
        >>> cartesian_prod([1, 2], [3, 1], [1, 4, 2])
        [[1, 3, 1], [1, 3, 4], [1, 3, 2], [1, 1, 1], [1, 1, 4], [1, 1, 2], [2, 3, 1], [2, 3, 4], [2, 3, 2], [2, 1, 1], [2, 1, 4], [2, 1, 2]]
    """
    if not isinstance(x, Iterable): x = [x]
    if len(y) == 0: return [[i] for i in x]
    res = []
    for i in x:
        res.extend([[i] + s for s in cartesian_prod(*y)])
    return res

def argmin(y, x=None):
    """
    Find the indices of minimal element in `y` given index domain `x`.
    
    Examples::
        >>> argmin([0, 2, 1, 4, 2], [1, 3, 4])
        [1, 4]
    """
    if x is None:
        if isinstance(y, dict): x = list(y.keys())
        else: x = range(len(y))
    if len(x) <= 0: return []
    m = min([y[i] for i in x])
    return [i for i in x if y[i] == m]

def argmax(y, x=None):
    """
    Find the indices of maximal element in `y` given index domain `x`.
    
    Examples::
        >>> argmin([0, 2, 1, 4, 2], [1, 3, 4])
        [3]
    """
    if x is None:
        if isinstance(y, dict): x = list(y.keys())
        else: x = range(len(y))
    if len(x) <= 0: return []
    m = max([y[i] for i in x])
    return [i for i in x if y[i] == m]

def min_argmin(y, x=None):
    """
    Find the minimal value as well as the indices in `y` given domain `x`.
    
    Examples::
        >>> min_argmin([0, 2, 1, 4, 2], [1, 3, 4])
        (2, [1, 4])
    """
    if x is None:
        if isinstance(y, dict): x = list(y.keys())
        else: x = range(len(y))
    if len(x) <= 0: return None, []
    m = min([y[i] for i in x])
    return m, [i for i in x if y[i] == m]

def max_argmax(y, x=None):
    """
    Find the maximal value as well as the indices in `y` given domain `x`.
    
    Examples::
        >>> argmin([0, 2, 1, 4, 2], [1, 3, 4])
        (4, [3])
    """
    if x is None:
        if isinstance(y, dict): x = list(y.keys())
        else: x = range(len(y))
    if len(x) <= 0: return None, []
    m = max([y[i] for i in x])
    return m, [i for i in x if y[i] == m]

def kth_biggest(list_, k: int):
    """
    Find the k-th biggest element in the list.
    
    Examples::
        >>> kth_biggest([0, 2, 1, 4, 2], 3)
        2
    """
    n = len(list_)
    avouch(k <= n, f"Cannot find the {k}-th element in a list of length {n}.")
    if n <= 1: return list_[0]
    p = randint(0, n - 1)
    lp = list_[p]
    gt_p = [x for x in list_ if x > lp]
    lt_p = [x for x in list_ if x < lp]
    n_gt_p = len(gt_p)
    n_lt_p = len(lt_p)
    if k <= n_gt_p: return kth_biggest(gt_p, k)
    elif n_gt_p < k <= n - n_lt_p: return lp
    else: return kth_biggest(lt_p, k - n + n_lt_p)
    
def kth_smallest(list_, k: int):
    n = len(list_)
    return kth_biggest(list_, n - k + 1)
    
def median(list_):
    n = len(list_)
    if n % 2 == 1: return kth_biggest(list_, n // 2)
    return (kth_biggest(list_, n // 2) + kth_biggest(list_, n // 2 + 1)) / 2

def flatten_list(list_):
    """
    Flat the nested lists `list_`.
    
    Examples::
        >>> flatten_list([0, 2, [1, 4, 2], [1, 3, 4]])
        [0, 2, 1, 4, 2, 1, 3, 4]
    """
    # Deprecated realization of the function, as elements may be strings with characters '[' or ']'.
    # items = str(list_).replace('[', '').replace(']', '').split(',')
    # return list(eval(','.join([x for x in items if x.strip() != ''])))
    flattened = []
    for x in list_:
        if isinstance(x, list):
            flattened.extend(flatten_list(x))
        else: flattened.append(x)
    return flattened

def item(list_):
    """
    Assert if the length of the list/tuple/set `list_` is not 1 and return the only element. 
    
    Examples::
        >>> item([0])
        0
        >>> item([1,2])
        AssertError: ...
    """
    list_ = to_list(list_)
    avouch(len(list_) == 1, f"Failure in itemize as the length of {repr(list_)} is not 1. ")
    return list_[0]

def to_list(x, l = None):
    """
    Try to cast element `x` into a list
    
    Examples::
        >>> to_list(1)
        [1]
        >>> to_list(0, 4)
        [0, 0, 0, 0]
        >>> to_list((1,2))
        [1, 2]
        >>> to_list((1,2), 4)
        [1, 2, 1, 2]
    """
    func_candidates = ['tolist', 'to_list', 'aslist', 'as_list', '__list__']
    for fname in func_candidates:
        if hasattr(x, fname) and callable(getattr(x, fname)):
            x = getattr(x, fname)(); break
    if isinstance(x, Iterable) and not isinstance(x, str): x = list(x)
    if not isinstance(x, list): x = [x]
    if l is None: return x
    if l % len(x) == 0: return x * (l // len(x))
    raise TypeError(f"{x} can not be converted into a list of length {l}. ")

def to_tuple(x, l = None):
    """
    Try to cast element `x` into a tuple of length `l`
    
    Examples::
        >>> to_tuple(1)
        (1,)
        >>> to_tuple(0, 4)
        (0, 0, 0, 0)
        >>> to_tuple([1,2])
        (1, 2)
        >>> to_tuple([1,2], 4)
        (1, 2, 1, 2)
    """
    func_candidates = ['totuple', 'to_tuple', 'astuple', 'as_tuple', '__tuple__']
    for fname in func_candidates:
        if hasattr(x, fname) and callable(getattr(x, fname)):
            x = getattr(x, fname)(); break
    try:
        return tuple(to_list(x, l))
    except TypeError:
        raise TypeError(f"{x} can not be converted into a tuple of length {l}. ")

def to_set(x):
    """
    Try to cast element `x` into a set
    
    Examples::
        >>> to_set(0)
        {0}
        >>> to_set([1,2])
        {1,2}
    """
    func_candidates = ['toset', 'to_set', 'asset', 'as_set', '__set__']
    for fname in func_candidates:
        if hasattr(x, fname) and callable(getattr(x, fname)): return getattr(x, fname)()
    return touch(lambda: set(x), touch(lambda: set(to_list(x)), {x}))

def map_ele(func, list_, index_ = None):
    """
    In-place! Map elements in `list_` at indices `index_` by function `func`. 
    
    Examples::
        >>> map_ele(lambda x: x+1, [1,2], 1)
        [1, 3]
        >>> map_ele(to_list, [1,2,3,4], [1,2])
        [1, [2], [3], 4]
    """
    if index_ is None: index_ = range(len(list_))
    if not index_: return list_
    index_ = to_list(index_)
    for i in index_: list_[i] = func(list_[i])
    return list_

def sublist(list_: list, index_):
    """
    Return elements in `list_` at indices `index_`. 
    
    Examples::
        >>> map_ele([1,2], [1])
        [2]
        >>> map_ele([1,2,3,4], [1,2])
        [2, 3]
    """
    if isinstance(index_, slice): index_ = range(index_.start, index_.stop)
    return [list_[i] for i in index_]

def arg_extract(arg:tuple, arg_type=None):
    """
    For *args, extract the only element if length is 1. 
    Set kwarg arg_type to define the types of objects that can be extracted. 
    By default, arg_type = None, which means any single element will be extracted. 
    
    Examples::
        >>> def f(*args): print(args)
        ... 
        >>> f([1,2,3,4])
        ([1, 2, 3, 4],)
        >>> def f(*args): print(arg_extract(args))
        ... 
        >>> f([1,2,3,4])
        [1, 2, 3, 4]
    """
    if len(arg) == 0: return ()
    if len(arg) > 1: arg = tuple(arg); return arg
    if arg_type is None: return arg[0]
    if not isinstance(arg_type, tuple): arg_type = (arg_type,)
    arg_type = tuple(t if isinstance(t, type) else t.__class__ for t in arg_type)
    if isinstance(arg[0], arg_type): arg = arg[0]
    return arg

def arg_tuple(arg:tuple, no_list=False):
    """
    Return the raw tuple. 
    
    Examples::
        >>> def f(*args): print(args)
        ... 
        >>> f([1,2,3,4])
        ([1, 2, 3, 4],)
        >>> def f(*args): print(arg_tuple(args))
        ... 
        >>> f([1,2,3,4])
        (1, 2, 3, 4)
    """
    if len(arg) == 1 and isinstance(arg[0], tuple): arg = arg[0]
    if len(arg) == 1 and isinstance(arg[0], list) and not no_list: arg = arg[0]
    arg = tuple(arg)
    return arg

def count(list_:list, filter):
    """
    Return number of elements in `list_` that satisfies `filter`. 
    
    Examples::
        >>> count([1,2], lambda x: x > 1)
        1
    """
    count = 0
    for x in list_:
        if filter(x): count += 1
    return count

def unique(list_:list):
    """
    Return a list of unique element in `list_`
    
    Examples::
        >>> count([1, 1, 3, 2])
        [1, 3, 2]
    """
    ulist = []
    for x in list_:
        if x not in ulist: ulist.append(x)
    return ulist

class infinite_itemize:
    def __init__(self, obj, n=None):
        if n is None or n < 0: n = None
        self.obj = obj
        self.num = n
        
    def __getitem__(self, i):
        if self.num is None or -self.num <= i < self.num: return self.obj
        raise IndexError("list index out of range")
    
    def __len__(self): return self.num
    
    def __str__(self): return f"infinite_itemize[{self.obj}]"
    
    @property
    def is_infinite(self): return self.num is None

def cat_generator(*gens):
    for gen in gens:
        for g in gen: yield g

def param_join(*models):
    return cat_generator(m.parameters() for m in models)




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2022-02",
    fileinfo = "File to execute shell commands for `zxhtools`.",
    requires = "pycamia"
).check()

__all__ = """
    AFF FFD TRS
""".split()

import re, os, sys

with __info__:
    from pycamia import Path, get_environ_vars

class TOOLS:
    def __init__(self, homeDIR='', verbose=False):
        self.delete = "rm -r" if 'darwin' in sys.platform.lower() or 'linux' in sys.platform.lower() else "del"
        self.verbose = verbose
        if not homeDIR: return
        self.toolsList = []
        for file in os.listdir(homeDIR):
            if os.path.isdir(file): continue
            self.toolsList.append(file.split('.')[0])
            self.__dict__[file.split('.')[0]] = os.path.join(homeDIR, file)

    def __getattr__(self, string):
        def run(x, variables={}, force=False):
            self(self.__dict__.get(string, string) + x, variables, force)
        return run

    def __call__(self, string, *args, **variables):
        force = False
        if len(args) >= 2: raise SyntaxError("Too much parameters for shell exchange! ")
        if 'force' in variables: force = variables['force']; variables.pop('force')
        if len(args) > 0: assert len(variables) == 0; variables = args[0]
        if 'variables' in variables: variables = variables['variables']
        cmd = string.format(**variables)
        if self.verbose: print("Running:", cmd)
        tmpcmd = string.format(**{k: v.replace(' ', '\ ') if isinstance(v, str) else v for k, v in variables.items()})
        do, *args = re.sub(r'[^\\] ', lambda x: x.group().replace(' ', '\n'), tmpcmd).split('\n')
        if self.__dict__.get(do, do) != do: return self(cmd.replace(do, self.__dict__.get(do, do)))
        for i, arg in enumerate(args):
            arg = arg.strip()
            if os.path.sep in arg:
                opt = args[i-1].strip()
                arg = arg.replace('\ ', ' ')
                arg = arg.strip('\'"')
                if not os.path.exists(arg):
                    if opt != '-o' and os.path.sep in ''.join(args[i + 1:]):
                        print('\n', cmd, sep='')
                        if not force: raise AssertionError("Path doesn't exists: " + arg)
                        else: print("Warning: Path doesn't exists: " + arg)
                    elif not os.path.exists(os.path.dirname(arg)): Path(arg).parent.mkdir()
                elif opt == '-o' and not force: return
        if self.verbose: os.system(cmd)
        else: return os.popen(cmd).read()

shell = TOOLS()

class tmpFile:
    def __init__(self, *path, preserve=False): self.file_list = list(path); self.update(path); self._p = preserve
    def __call__(self, *path): self.file_list.extend(path); self.update(path); return path[0]
    def __del__(self): [os.remove(f) for f in self.file_list if os.path.exists(f) and not self._p]
    def update(self, flist): [Path(f).mkdir() for f in flist]

class trans_cmd(str):
    def __new__(cls, string, *transformations, source = None, target = None, verbose = False):
        if "zxhtransform" in string: return super().__new__(cls, string)
        assert source and target
        transformations = (string,) + transformations
        cmd = ["zxhtransform", target, "{source}", "-o {outfilename}", "-n", str(len(transformations))]
        for t in transformations: cmd.extend(['-t', t])
        if not verbose: cmd.append("-v 0")
        self = super().__new__(cls, ' '.join(cmd))
        self.target_space = (source, target)
        return self
        
    def __add__(x, y): return trans_cmd(str(x).rstrip(' ') + ' ' + y.lstrip(' '))
    def __invert__(self):
        _, target, source, _, outfilename, *_ = self.split()
        target = self.target_space[(self.target_space.index(target) + 1) % 2]
        transformations = [x.strip() for x in re.findall(r'-t ([^ ]+)', self)]
        options = [x.strip() for x in re.findall(r' (-[^-]+)', self) if x[1] not in "not"]
        options += [x.strip() for x in re.findall(r' (--[^-]+)', self)]
        newcmd = ["zxhtransform", target, source, "-o", outfilename, "-n", str(len(transformations))]
        for t in transformations[::-1]: newcmd.extend(['-t', t])
        if '-forward' in options:
            while '-forward' in options: options.remove('-forward')
        else: options.append('-forward')
        newcmd.extend(options)
        return trans_cmd(' '.join(newcmd))

    def __call__(self, fin, fout, **kwargs):
        shell(self, source = fin, outfilename = fout, **kwargs)

    def transformations(self, i = None):
        T = [x.strip() for x in re.findall(r'-t ([^ ]+)', self)]
        if i is None: return T
        else: return T[i]
    def replace(self, *args): return trans_cmd(super().replace(*args))

def registration(code, target, source, output = '', 
    save_images = False, save_transforms = True, variables = {}, 
    verbose = False, trans_only = False, reg_file_format = lambda s: os.path.join('reg_files', s)):
    '''
    code with format:
    rigid -Reg 2 -sub 10 10 10 -sub 8 8 8 -length 3 1.5 -steps 200 200
    affine -Reg 2 -sub 8 8 8 -sub 6 6 6 -length 1 0.5 -steps 200 50
    FFD -Reg 3 -ffd 80 80 80 -sub 8 8 8 -ffd 40 40 40 -sub 6 6 6 -ffd 20 20 20 -sub 6 6 6 -length 2 2 1 -steps 200 -bending 0.0001
    '''
    shell = TOOLS(verbose = verbose)
    lines = [l.strip() for l in code.split('\n') if l.strip()]
    outfile = '', 'NONE'
    name, ext = 0, 1
    operations = {'rigid': 'AFF', 'affine': 'AFF', 'ffd': 'FFD'}
    header = {'rigid': "zxhreg -rig", 'affine': "zxhregaff -aff", 'ffd': "zxhregffdsemi0"}
    new_source = reg_file_format(Path(source) - Path(source) ** Path(target))
    sourcefile = new_source
    tmp_images = tmpFile(preserve = save_images)
    tmp_transforms = tmpFile(preserve = save_transforms)
    transformations = []
    for i, line in enumerate(lines):
        do = line.split()[0]
        if outfile[ext] == 'NONE': pass
        elif outfile[ext] == operations[do.lower()]:
            transformations.pop(-1)
        else: sourcefile = outfile[name] + '.nii.gz'
        outfile = '.'.join((sourcefile, do.lower())), operations[do.lower()]
        transformations.append('.'.join(outfile))
    sourcefile = new_source
    outfile = '', 'NONE'
    if trans_only:
        for t in transformations:
            if not os.path.isfile(t):
                trans_only = False
                break
    for i, line in enumerate(lines):
        do = line.split()[0]
        s = line.find(do)
        options = (line[:s] + line[s+len(do):]).strip()
        if outfile[ext] == 'NONE':
            if '-pre' not in line: prealign = '-pre 13'
            else:
                prealign = re.findall(r' +-pre +[1-9]+| +-pre +0 +[^ ]+', options)[-1]
                options = options.replace(prealign, '')
                prealign = prealign.strip()
        elif outfile[ext] == operations[do.lower()]:
            prealign = '-pre 0 ' + '.'.join(outfile)
        else: prealign = ''
        if not prealign: sourcefile = outfile[name] + '.nii.gz'
        outfile = '.'.join((sourcefile, do.lower())), operations[do.lower()]
        tmp_images(outfile[name] + '.nii.gz')
        tmp_transforms('.'.join(outfile))
        if sourcefile == new_source: sfile = source
        else: sfile = sourcefile
        cmd = ' '.join((header[do.lower()], "-target", target,
                        "-source %s -o %s"%(sfile, outfile[name]),
                        options, prealign)).strip()
        if i >= len(lines) - 1 or operations[lines[i+1].split()[0].lower()] != outfile[ext]: pass
        else: cmd += " -notsaveimage"
        if not verbose: cmd += " -v 0"
        if not trans_only:
            shell(cmd, variables, force = True)
            if i >= len(lines) - 1 and output: shell("cp %s %s"%(outfile[name] + '.nii.gz', output))
    return trans_cmd(*transformations[::-1], source = source, target = target, verbose = verbose)

if __name__ == '__main__': pass




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2022-01",
    fileinfo = "File of numpy operations. ",
    requires = 'numpy'
)

__all__ = """
    toU
    toI
""".split()

try:
    with __info__:
        import numpy as np
    
    def toU(dt):
        """
        Convert a numpy dtype `dt` to unsigned version. 
        """
        if hasattr(dt, 'dtype'): x = dt; dt = dt.dtype
        dt = np.dtype(str(dt).split('.')[-1])
        if dt.kind == np.dtype(np.int16).kind: rdt = np.dtype('uint%d'%(8*dt.itemsize))
        else: rdt = dt
        if isinstance(x, np.ndarray): return x.astype(rdt)
        
    def toI(dt):
        """
        Convert a numpy dtype `dt` to signed version. 
        """
        if hasattr(dt, 'dtype'): x = dt; dt = dt.dtype
        dt = np.dtype(str(dt).split('.')[-1])
        if dt.kind == np.dtype(np.uint).kind: rdt = np.dtype('int%d'%(8*dt.itemsize))
        else: rdt = dt
        if isinstance(x, np.ndarray): return x.astype(rdt)

except: pass
